# SPDX-License-Identifier: MIT
"""Action base and specific classes"""

from . import defaults
from logging import getLogger, DEBUG, INFO, WARNING, ERROR
from tornado.httpclient import HTTPClient
from urllib.parse import urlencode
from smtplib import SMTP_SSL
from email.mime.text import MIMEText
from email.utils import make_msgid, formatdate
from subprocess import run
import json

import ssl

_log = getLogger('fletchck.action')
_log.setLevel(INFO)

ACTION_TYPES = {}

_PATCH_TYPES = {
    'cksms': 'sms',
}


def loadAction(name, config):
    """Return an action object for the provided config dict"""
    ret = None
    if config['type'] in _PATCH_TYPES:
        config['type'] = _PATCH_TYPES[config['type']]
    if config['type'] in ACTION_TYPES:
        options = defaults.getOpt('options', config, dict, {})
        ret = ACTION_TYPES[config['type']](name, options)
        ret.actionType = config['type']
    else:
        _log.warning('%s: Invalid action type ignored', name)
    return ret


class BaseAction():
    """Action base class, implements the log type and interface"""

    def __init__(self, name=None, options={}):
        self.name = name
        self.options = options
        self.actionType = 'log'

    def getStrOpt(self, key, default=None):
        return defaults.getOpt(key, self.options, str, default)

    def getIntOpt(self, key, default=None):
        return defaults.getOpt(key, self.options, int, default)

    def getListOpt(self, key, default=None):
        return defaults.getOpt(key, self.options, list, default)

    def getBoolOpt(self, key, default=None):
        return defaults.getOpt(key, self.options, bool, default)

    def _notify(self, source):
        _log.warning('%s (%s): %s', source.name, source.checkType,
                     source.getState())
        return True

    def trigger(self, source):
        count = 0
        while True:
            if self._notify(source):
                break
            count += 1
            if count >= defaults.ACTIONTRIES:
                _log.error('%s (%s): Fail after %r tries', self.name,
                           self.actionType, count)
                return False
        return True

    def flatten(self):
        """Return the action detail as a flattened dictionary"""
        optMap = {}
        for key in self.options:
            optMap[key] = self.options[key]
        return {'type': self.actionType, 'options': optMap}


class sendEmail(BaseAction):
    """Send email by configured submit"""

    def _notify(self, source):
        site = self.getStrOpt('site', defaults.APPNAME)

        subject = "[%s] %s (%s) in %s state" % (
            site, source.name, source.checkType, source.getState())
        ml = []
        ml.append('%s (%s) in %s state at %s%s' %
                  (source.name, source.checkType, source.getState(),
                   source.lastFail if source.failState else source.lastPass,
                   '' if source.failState else '\n\U0001f436\U0001F44D'))
        if source.log:
            ml.append('')
            ml.append('Log:')
            ml.append('')
            for l in source.log:
                ml.append(l)
        message = '\n'.join(ml)
        username = self.getStrOpt('username')
        password = self.getStrOpt('password')
        sender = self.getStrOpt('sender')
        recipients = self.getListOpt('recipients', [])
        hostname = self.getStrOpt('hostname')
        fallback = self.getStrOpt('fallback')
        port = self.getIntOpt('port', 0)
        timeout = self.getIntOpt('timeout', defaults.SUBMITTIMEOUT)

        _log.debug('Send email to %r via %r : %r', recipients, hostname,
                   subject)

        ret = True
        if not recipients:
            _log.warning('No email recipients specified - notify ignored')
            return ret

        msgid = make_msgid()
        m = MIMEText(message)
        if sender:
            m['From'] = sender
        m['Subject'] = subject
        m['Message-ID'] = msgid
        m['Date'] = formatdate(localtime=True)

        if hostname:
            ret = False
            try:
                ctx = ssl.create_default_context()
                with SMTP_SSL(host=hostname,
                              port=port,
                              timeout=timeout,
                              context=ctx) as s:
                    if username and password:
                        s.login(username, password)
                    s.send_message(m, from_addr=sender, to_addrs=recipients)
                ret = True
            except Exception as e:
                _log.warning('SMTP Email Notify failed: %s', e)
        elif fallback:
            ret = False
            try:
                cmd = [fallback, '-oi']
                if sender:
                    cmd.extend(['-r', sender])
                cmd.append('--')
                cmd.extend(recipients)
                run(cmd,
                    capture_output=True,
                    input=m.as_bytes(),
                    timeout=timeout,
                    check=True)
                _log.debug('Fallback email sent ok to: %r', recipients)
                ret = True
            except Exception as e:
                _log.warning('Fallback Email Notify failed: %s', e)
        else:
            _log.warning('Email notify not configured')
        return ret


class ckSms(BaseAction):
    """Post SMS via cloudkinnect api"""

    def trigger(self, source):
        message = '%s: %s\n%s\n%s'
        if source.failState:
            message = '\U0001f436\U0001f4ac\n%s: %s\n%s\n%s' % (
                source.name,
                source.getState(),
                source.getSummary(),
                source.lastFail,
            )
        else:
            message = '\U0001f436\U0001F44D\n%s: %s\n%s' % (
                source.name,
                source.getState(),
                source.lastPass,
            )
        sender = self.getStrOpt('sender', 'dedicated')
        recipients = [i for i in self.getListOpt('recipients', [])]
        url = self.getStrOpt('url', defaults.CKURL)
        apikey = self.getStrOpt('apikey')

        httpClient = HTTPClient()
        failCount = 0
        while recipients and failCount < defaults.ACTIONTRIES:
            recipient = ','.join(recipients).replace('+', '')
            _log.debug('Send sms to %r via %r : %r', recipient, url, message)
            postBody = urlencode({
                'originator': sender,
                'mobile_number': recipient,
                'concatenated': 'true',
                'utf': 'true',
                'text': message
            })
            try:
                response = httpClient.fetch(
                    url,
                    method='POST',
                    headers={
                        'Authorization': 'Bearer ' + apikey,
                        'Content-Type': 'application/x-www-form-urlencoded'
                    },
                    body=postBody)
                if response.code == 200:
                    recipients = None
                else:
                    failCount += 1
                    _log.warning('SMS Notify failed: %r:%r', response.code,
                                 response.body)
            except Exception as e:
                failCount += 1
                _log.warning('SMS Notify failed: %s', e)
        return not recipients


ACTION_TYPES['email'] = sendEmail
ACTION_TYPES['sms'] = ckSms
ACTION_TYPES['log'] = BaseAction



# SPDX-License-Identifier: MIT
"""Application support utilities"""

import os
import re
import sys
import csv
import json
import struct
import math
from datetime import datetime, timedelta, timezone
# Python < 3.11 workaround for datetime.UTC
UTC = timezone.utc
from secrets import randbits, token_hex
from passlib.hash import argon2 as kdf
from tempfile import NamedTemporaryFile, mkdtemp
from logging import getLogger, Handler, DEBUG, INFO, WARNING
from subprocess import run
from ipaddress import IPv6Address
from . import defaults
from .action import loadAction
from .check import loadCheck, getZone, BaseCheck, timeString
from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.interval import IntervalTrigger
from apscheduler.triggers.cron import CronTrigger

_log = getLogger('fletchck.util')
_log.setLevel(DEBUG)
getLogger('apscheduler.executors').setLevel(INFO)
getLogger('apscheduler.executors.default').setLevel(INFO)

_INTTRIGKEYS = {
    'weeks',
    'days',
    'hours',
    'minutes',
    'seconds',
}
_INTERVALKEYS = {
    'weeks': 'week',
    'days': 'day',
    'hours': 'hr',
    'minutes': 'min',
    'seconds': 'sec',
    'end_date': 'end',
    'timezone': 'z',
}
_CRONKEYS = {
    'year': 'year',
    'month': 'month',
    'day': 'day',
    'week': 'week',
    'day_of_week': 'weekday',
    'hour': 'hr',
    'minute': 'min',
    'second': 'sec',
    'end_date': 'end',
    'timezone': 'z',
}
_BOOLVALS = {
    't': True,
    'f': False,
    '1': True,
    '0': False,
    'true': True,
    'false': False,
    'yes': True,
    'no': False,
    'null': False,
    'none': False,
    '': False,
}


class SaveFile():
    """Tempfile-backed save file contextmanager.

       Creates a temporary file with the desired mode and encoding
       and returns a context manager and writable file handle.

       On close, the temp file is atomically moved to the provided
       filename (if possible).
    """

    def __init__(self,
                 filename,
                 mode='t',
                 encoding='utf-8',
                 tempdir='.',
                 perm=0o600):
        self.__sfile = filename
        self.__path = tempdir
        self.__perm = perm
        if mode == 'b':
            encoding = None
        self.__tfile = NamedTemporaryFile(mode='w' + mode,
                                          suffix='.tmp',
                                          prefix='sav_',
                                          dir=self.__path,
                                          encoding=encoding,
                                          delete=False)

    def __enter__(self):
        return self.__tfile

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.__tfile.close()
        if exc_type is not None:
            return False  # raise exception
        # otherwise, file is saved ok in temp file
        os.chmod(self.__tfile.name, self.__perm)
        os.rename(self.__tfile.name, self.__sfile)
        return True


class LogHandler(Handler):

    def __init__(self, site):
        self.site = site
        Handler.__init__(self)

    def emit(self, record):
        """Append record to log and prune early entries"""
        msg = self.format(record)
        self.site.log.append(msg)
        if len(self.site.log) > 200:
            del (self.site.log[0:10])


def toBool(boolstr):
    key = boolstr.lower()
    if key in _BOOLVALS:
        return _BOOLVALS[key]
    if key.isdigit():
        return bool(int(key))
    else:
        return bool(key)


def toHostname(hostname):
    """Check hostname for MAC or unscoped ipv6 ll"""
    if isMacAddr(hostname):
        # assume EUI-64 LL address desired
        hostname = mac2ll(hostname)
    elif ':' in hostname:
        # assume ipv6 - check for LL without scope
        hostname = lladdscope(hostname)
    else:
        # probably hostname or v4
        pass
    return hostname


def isMacAddr(hostname):
    """Return true if hostname looks like a MAC address"""
    return re.match('^[\dA-Fa-f]{2}(:[\dA-Fa-f]{2}){5}$', hostname) is not None


def lladdscope(address):
    """Append interface 2 to LL address if scope not provided"""
    ret = address
    try:
        va = IPv6Address(address)
        if va.is_link_local and va.scope_id is None:
            ret = '%s%%2' % (va, )
    except Exception as e:
        _log.info('Invalid ipv6 address %s: %s', e.__class__.__name__, e)
    return ret


def mac2ll(macaddr, interface='2'):
    """Convert MAC address to EUI-64 link-local addr on interface 2"""
    ret = macaddr
    try:
        if isMacAddr(macaddr):
            leui48 = int(macaddr.replace(':', ''), 16) | 1 << 41
            llval = 0xfe80 << 112 | 0xfffe << 24 | (
                leui48 & 0xffffff000000) << 16 | leui48 & 0xffffff
            ##addr = IPv6Address(llval)
            ret = '%s%%%s' % (IPv6Address(llval), interface)
    except Exception as e:
        _log.info('Invalid MAC address %s: %s', e.__class__.__name__, e)
    return ret


def trigger2Text(trigger):
    """Convert a trigger schedule object to text string"""
    rv = []
    if isinstance(trigger, dict):
        if 'interval' in trigger:
            rv.append('interval')
            for key in _INTERVALKEYS:
                if key in trigger['interval']:
                    rv.append(str(trigger['interval'][key]))
                    rv.append(_INTERVALKEYS[key])
        elif 'cron' in trigger:
            rv.append('cron')
            for key in _CRONKEYS:
                if key in trigger['cron']:
                    rv.append(str(trigger['cron'][key]))
                    rv.append(_CRONKEYS[key])
    return ' '.join(rv)


def text2Trigger(triggerText):
    """Read, validate and return trigger definition"""
    ret = None
    try:
        trigger = None
        if triggerText:
            tv = triggerText.lower().split()
            if tv:
                # check type prefix
                trigType = tv[0]
                if trigType == 'interval':
                    tv.pop(0)
                    trigger = {'interval': {}}
                elif trigType == 'cron':
                    tv.pop(0)
                    trigger = {'cron': {}}
                else:
                    _log.debug('text2Trigger: Assuming interval for %r', tv)
                    trigger = {'interval': {}}

                keyMap = {}
                trigMap = None
                if 'interval' in trigger:
                    trigMap = trigger['interval']
                    for k in _INTERVALKEYS:
                        keyMap[k] = k
                        keyMap[_INTERVALKEYS[k]] = k
                elif 'cron' in trigger:
                    trigMap = trigger['cron']
                    for k in _CRONKEYS:
                        keyMap[k] = k
                        keyMap[_CRONKEYS[k]] = k

                # scan input text
                nextVal = []
                while tv:
                    # check for value:
                    if tv[0] in keyMap:
                        _log.debug('Ignoring spurious unit %s', tv[0])
                        tv.pop(0)
                        continue

                    nextVal.append(tv.pop(0))
                    if tv:
                        if tv[0] in keyMap:
                            unit = tv.pop(0)
                            val = ' '.join(nextVal)
                            key = keyMap[unit]
                            if key in _INTTRIGKEYS:
                                val = int(val)
                            if key in trigMap:
                                _log.debug('Trigger key %s re-defined', key)
                            trigMap[key] = val
                            nextVal = []
                if nextVal:
                    # Lazily assume minutes for degenerate input
                    val = ' '.join(nextVal)
                    _log.debug(
                        'Extra value without units %s, assuming minutes', val)
                    key = keyMap['min']
                    if key in _INTTRIGKEYS:
                        val = int(val)
                    if key in trigMap:
                        _log.debug('Trigger key %s re-defined', key)
                    trigMap[key] = val
                    nextVal = []

                # try and create a trigger from the definition
                if 'interval' in trigger:
                    t = IntervalTrigger(**trigMap)
                elif 'cron' in trigger:
                    t = CronTrigger(**trigMap)

                ret = trigger
    except Exception as e:
        _log.info('Invalid trigger %s: %s', e.__class__.__name__, e)
    return ret


def checkPass(pw, hash):
    return kdf.verify(pw, hash[0:1024])


def createHash(pw):
    return kdf.using(rounds=defaults.PASSROUNDS).hash(pw)


def randPass():
    """Return a random passkey"""
    choiceLen = len(defaults.PASSCHARS)
    if choiceLen < 8:
        raise RuntimeError('Unexpected length passchars')
    depth = int(math.floor(math.log2(choiceLen)))
    clen = 2**depth
    if clen != choiceLen:
        _log.warning('Using first %r chars of passchars', clen)
    passLen = int(math.ceil(defaults.PASSBITS / depth))
    rawBits = randbits(passLen * depth)
    mask = clen - 1
    pv = []
    for i in range(0, passLen):
        pv.append(defaults.PASSCHARS[rawBits & mask])
        rawBits >>= depth
    return ''.join(pv)


def saveSite(site):
    """Save the current site state to disk"""
    dstCfg = {'base': site.base, 'timezone': None, 'webui': None}
    if site.timezone is not None:
        dstCfg['timezone'] = site.timezone.key
    if site.webCfg is not None:
        dstCfg['webui'] = {}
        for k in defaults.WEBUICONFIG:
            dstCfg['webui'][k] = site.webCfg[k]
    if site.mqttCfg is not None:
        dstCfg['mqtt'] = {}
        for k in defaults.MQTTCONFIG:
            dstCfg['mqtt'][k] = site.mqttCfg[k]
    dstCfg['actions'] = {}
    for a in site.actions:
        dstCfg['actions'][a] = site.actions[a].flatten()
    dstCfg['checks'] = {}
    for c in site.checks:
        dstCfg['checks'][c] = site.checks[c].flatten()
    dstCfg['log'] = site.log

    # backup existing config and save
    tmpName = None
    if os.path.exists(site.configFile):
        tmpName = site.configFile + token_hex(6)
        os.link(site.configFile, tmpName)
    with SaveFile(site.configFile) as f:
        json.dump(dstCfg, f, indent=1)
    if tmpName is not None:
        os.rename(tmpName, site.configFile + '.bak')
    _log.debug('Saved site config to %r', site.configFile)


def readCsv(csvFile):
    """Read CSV file and return a check config."""
    checkOpts = (
        'trigger',
        'paused',
        'threshold',
        'retries',
        'priority',
        'failAction',
        'passAction',
        'publish',
        'remoteId',
        'timezone',
    )
    specialOpts = (
        'name',
        'checkType',
        'depends',
        'actions',
        'checks',
        'inseq',
    )
    optionTypes = {
        'trigger': text2Trigger,
        'paused': toBool,
        'threshold': int,
        'retries': int,
        'priority': int,
        'failAction': toBool,
        'passAction': toBool,
        'timezone': getZone,
        'hostname': toHostname,
        'port': int,
        'tls': toBool,
        'selfsigned': toBool,
        'level': int,
        'timeout': int,
        'temperature': int,
        'reqName': str,
        'reqPath': str,
        'reqType': str,
        'probe': str,
        'hostkey': str,
        'serialPort': str,
        'volume': str,
        'reqTcp': toBool,
        'beeper': toBool,
        'hysteresis': int,
    }
    inseq = {}
    hdr = None
    dat = []
    cfg = {'checks': {}}
    cr = csv.reader(csvFile)

    # read in csv rows to intermediate key:value with converted values
    for r in cr:
        if hdr is None:
            # assume first row is header
            hdr = [c.strip() for c in r]
            if 'name' not in hdr or 'checkType' not in hdr:
                _log.error('CSV missing required columns: name, checkType')
                raise RuntimeError(
                    'CSV missing required headers: name, checkType')
        else:
            count = 0
            chk = {}
            for c in r:
                c = c.strip()
                if c != '':
                    key = hdr[count]
                    val = c
                    if key in optionTypes:
                        val = optionTypes[key](c)
                    chk[key] = val
                count += 1
            if chk['checkType'] == 'sequence':
                inseq[chk['name']] = []
            dat.append(chk)

    # convert intermediate options into check configs
    for chk in dat:
        checkName = chk['name']
        checkType = chk['checkType']
        checkCfg = {"type": checkType}
        for key in checkOpts:
            if key in chk:
                checkCfg[key] = chk[key]
        if 'depends' in chk:
            checkCfg['depends'] = []
            for dep in chk['depends'].split(','):
                dep = dep.strip()
                if dep:
                    checkCfg['depends'].append(dep)
        if 'actions' in chk:
            checkCfg['actions'] = []
            for dep in chk['actions'].split(','):
                dep = dep.strip()
                if dep:
                    checkCfg['actions'].append(dep)
        checkCfg['options'] = {}
        for key in chk:
            if key not in specialOpts and key not in checkOpts:
                checkCfg['options'][key] = chk[key]
        if 'checks' in chk:
            checkCfg['options']['checks'] = []
            for dep in chk['checks'].split(','):
                dep = dep.strip()
                if dep:
                    checkCfg['options']['checks'].append(dep)
        if 'inseq' in chk:
            checkCfg['inseq'] = []
            for dep in chk['inseq'].split(','):
                dep = dep.strip()
                if dep:
                    checkCfg['inseq'].append(dep)
        cfg['checks'][checkName] = checkCfg
    _log.debug('Read %d check entries from CSV file', len(cfg['checks']))
    return cfg


def mergeCsv(path, config, csvFilename):
    """Import checks from csv file and merge with site config"""
    importFilename = os.path.realpath(csvFilename)
    cfgFilename = os.path.join(path, config)
    if os.path.samefile(cfgFilename, importFilename):
        _log.warning('Ignored CSV import from existing config')
        return
    if os.path.exists(importFilename):
        _log.info('Importing checks from CSV %r', importFilename)
        try:
            importConf = None
            with open(importFilename) as f:
                importConf = readCsv(f)
            cfgConf = None
            with open(cfgFilename) as f:
                cfgConf = json.load(f)
            if mergeCfg(importConf, cfgConf):
                _log.info('Saving updated config')
                tmpName = None
                if os.path.exists(cfgFilename):
                    tmpName = cfgFilename + token_hex(6)
                    os.link(cfgFilename, tmpName)
                with SaveFile(cfgFilename) as f:
                    json.dump(cfgConf, f, indent=1)
                if tmpName is not None:
                    os.rename(tmpName, cfgFilename + '.bak')
            else:
                _log.warning('No updates imported')
        except Exception as e:
            _log.warning('Ignored invalid import, %s: %s',
                         e.__class__.__name__, e)
            raise e
    else:
        _log.warning('CSV checks import file not found, ignored')


def mergeCfg(importConf, cfgConf):
    """Copy values from importConf dict to cfgConf dict."""
    doSave = False
    if 'timezone' in importConf:
        _log.info('Imported timezone')
        cfgConf['timezone'] = importConf['timezone']
        doSave = True
    if 'webui' in importConf and importConf['webui'] is not None:
        if 'webui' not in cfgConf or not isinstance(cfgConf['webui'], dict):
            cfgConf['webui'] = {}
        dst = cfgConf['webui']
        src = importConf['webui']
        for key in src:
            if key == 'users':
                if 'users' not in dst or not isinstance(dst['users'], dict):
                    dst['users'] = {}
                for user in src['users']:
                    if user == defaults.ADMINUSER and defaults.ADMINUSER in dst[
                            'users']:
                        _log.warning('Admin user not updated')
                    else:
                        dst['users'][user] = src['users'][user]
            else:
                dst[key] = src[key]
        _log.info('Imported webui config')
        doSave = True
    if 'mqtt' in importConf:
        if 'mqtt' not in cfgConf:
            cfgConf['mqtt'] = {}
        dst = cfgConf['mqtt']
        src = importConf['mqtt']
        for key in src:
            dst[key] = src[key]
        _log.info('Imported mqtt config')
        doSave = True
    if 'actions' in importConf:
        for action in importConf['actions']:
            if action not in cfgConf['actions']:
                cfgConf['actions'][action] = {}
            destAction = cfgConf['actions'][action]
            srcAction = importConf['actions'][action]
            if 'type' in srcAction:
                destAction['type'] = srcAction['type']
            if 'options' in srcAction:
                if 'options' not in destAction:
                    destAction['options'] = {}
                for key in srcAction['options']:
                    destAction['options'][key] = srcAction['options'][key]
            _log.info('Imported action: %s', action)
            doSave = True
    inseq = {}
    if 'checks' in importConf:
        for checkName in importConf['checks']:
            if checkName not in cfgConf['checks']:
                cfgConf['checks'][checkName] = {}
            destCheck = cfgConf['checks'][checkName]
            srcCheck = importConf['checks'][checkName]
            for key in srcCheck:
                if key != 'data' and key != 'inseq':
                    destCheck[key] = srcCheck[key]
            if 'inseq' in srcCheck:
                for seqName in srcCheck['inseq']:
                    if seqName:
                        if seqName not in inseq:
                            inseq[seqName] = []
                    if checkName not in inseq[seqName]:
                        inseq[seqName].append(checkName)
            _log.info('Imported check: %s', checkName)
            doSave = True
    for seqName in inseq:
        if seqName in cfgConf['checks']:
            seq = cfgConf['checks'][seqName]
            if seq['type'] == 'sequence':
                if 'checks' not in seq['options']:
                    seq['options']['checks'] = []
                for checkName in inseq[seqName]:
                    if checkName not in seq['options']['checks']:
                        seq['options']['checks'].append(checkName)
                        _log.debug('Added %s to sequence %s', checkName,
                                   seqName)
                        doSave = True
            else:
                _log.warning('Named inseq not a sequence:  %s', seqName)
        else:
            _log.warning('Ignored unknown sequence %s', seqName)
    return doSave


def mergeConfig(path, config, option):
    """Merge selected values from option into config."""
    importFilename = os.path.realpath(option)
    cfgFilename = os.path.join(path, config)
    if os.path.samefile(cfgFilename, importFilename):
        _log.warning('Ignored import from existing config')
        return
    if os.path.exists(importFilename):
        _log.info('Importing config from %r', importFilename)
        try:
            importConf = None
            with open(importFilename) as f:
                importConf = json.load(f)
            cfgConf = None
            with open(cfgFilename) as f:
                cfgConf = json.load(f)
            if mergeCfg(importConf, cfgConf):
                _log.info('Saving updated config')
                tmpName = None
                if os.path.exists(cfgFilename):
                    tmpName = cfgFilename + token_hex(6)
                    os.link(cfgFilename, tmpName)
                with SaveFile(cfgFilename) as f:
                    json.dump(cfgConf, f, indent=1)
                if tmpName is not None:
                    os.rename(tmpName, cfgFilename + '.bak')
            else:
                _log.warning('No updates imported')

        except Exception as e:
            _log.warning('Ignored invalid import, %s: %s',
                         e.__class__.__name__, e)
    else:
        _log.warning('Import file not found, ignored')


def initSite(path, webUi=True, webPort=None):
    """Prepare a new empty site under path, returns True to continue"""
    cfgPath = os.path.realpath(path)
    cfgFile = os.path.join(cfgPath, defaults.CONFIGPATH)
    backup = False

    # check for an existing config
    if os.path.exists(cfgFile):
        choice = None
        if sys.stdin.isatty():
            prompt = 'Replace existing site? (y/N) '
            choice = input(prompt)
        if not choice or choice.lower()[0] != 'y':
            _log.error('Existing site not overwritten')
            return False

    # create initial configuration
    siteCfg = {}
    siteCfg['base'] = cfgPath
    siteCfg['timezone'] = defaults.TIMEZONE
    if webUi:
        siteCfg['webui'] = dict(defaults.WEBUICONFIG)
        if webPort is not None:
            siteCfg['webui']['por'] = max(min(webPort, 65535), 1)
        else:
            siteCfg['webui']['port'] = 30000 + randbits(15)
        mkCert(cfgPath, siteCfg['webui']['hostname'])
        siteCfg['webui']['cert'] = os.path.join(cfgPath, defaults.SSLCERT)
        siteCfg['webui']['key'] = os.path.join(cfgPath, defaults.SSLKEY)

        # create admin user
        siteCfg['webui']['users'] = {}
        adminPw = randPass()
        siteCfg['webui']['users'][defaults.ADMINUSER] = createHash(adminPw)
        # add dummy hash for unknown users
        siteCfg['webui']['users'][''] = createHash(randPass())
    else:
        siteCfg['webui'] = None

    # Add an empty set of checks
    siteCfg['checks'] = {}

    # saveconfig
    tmpName = None
    if os.path.exists(cfgFile):
        tmpName = cfgFile + token_hex(6)
        os.link(cfgFile, tmpName)
    with SaveFile(cfgFile) as f:
        json.dump(siteCfg, f, indent=1)
    if tmpName is not None:
        os.rename(tmpName, cfgFile + '.bak')

    # report
    if webUi:
        print(
            '\nSite address:\thttps://%s:%d\nAdmin password:\t%s\n' %
            (siteCfg['webui']['hostname'], siteCfg['webui']['port'], adminPw))
    else:
        print('\nConfigured without web interface.\n')
    if sys.stdin.isatty():
        choice = input('Start? (Y/n) ')
        if choice and choice.lower()[0] == 'n':
            return False
        else:
            return True
    return False


def updateCheck(site, oldName, newName, config):
    """Update an existing check on a running site"""
    # un-schedule
    job = site.scheduler.get_job(oldName)
    if job is not None:
        _log.debug('Removing %s (%r) from schedule', oldName, job)
        site.scheduler.remove_job(oldName)

    # fetch handle to old check and remove from site
    oldCheck = site.checks[oldName]
    del site.checks[oldName]

    # add updated config to site with new name
    addCheck(site, newName, config, update=True)

    # repair dependencies and sequences
    newCheck = site.checks[newName]
    for name in site.checks:
        if name != newName:
            c = site.checks[name]
            c.replace_depend(oldName, newCheck)
            if c.checkType == 'sequence':
                c.replace_check(oldName, newCheck)
            if oldName != newName:
                if 'checks' in c.options:
                    if isinstance(c.options['checks'], list):
                        cl = c.options['checks']
                        if oldName in cl:
                            cl[cl.index(oldName)] = newName
    # refresh site map
    site.checkMap(True)


def addAction(site, name, config):
    nAct = loadAction(name, config)
    if nAct is not None:
        site.actions[name] = nAct
        _log.warning('Added action %s to site', name)


def addCheck(site, name, config, update=False):
    """Add the named check to running site, or replace existing if update set"""
    newCheck = loadCheck(name, config, site.timezone)

    # add actions to check
    if 'actions' in config:
        if isinstance(config['actions'], list):
            for a in config['actions']:
                if a in site.actions:
                    newCheck.add_action(site.actions[a])
                else:
                    _log.info('%s ignored unknown action %s', name, a)

    # update check dependencies
    if 'depends' in config:
        if isinstance(config['depends'], list):
            for d in config['depends']:
                if d in site.checks:
                    newCheck.add_depend(site.checks[d])

    # update sequence checks
    if newCheck.checkType == 'sequence':
        if 'checks' in newCheck.options:
            if isinstance(newCheck.options['checks'], list):
                for subName in newCheck.options['checks']:
                    if subName in site.checks:
                        newCheck.add_check(site.checks[subName])

    # add check to site
    site.checks[name] = newCheck
    if newCheck.remoteId is not None:
        site.remotes[newCheck.remoteId] = name
    _log.debug('Load check %r (%s)', name, newCheck.checkType)

    # schedule check
    if newCheck.trigger is not None:
        trigOpts = None
        trigType = None
        if 'interval' in newCheck.trigger:
            trigType = 'interval'
            trigOpts = dict(newCheck.trigger['interval'])
            # set first trigger to run shortly after adding
            trigOpts['start_date'] = datetime.now(UTC) + timedelta(
                seconds=defaults.FIRSTTRIGGER)
        elif 'cron' in newCheck.trigger:
            trigType = 'cron'
            trigOpts = dict(newCheck.trigger['cron'])
        if trigType is not None:
            _log.debug('Adding %s %s trigger to schedule: %r', name, trigType,
                       trigOpts)
            if 'timezone' not in trigOpts:
                if newCheck.timezone:
                    trigOpts['timezone'] = newCheck.timezone
                elif site.timezone:
                    trigOpts['timezone'] = site.timezone
            trigOpts['jitter'] = defaults.TRIGGERJITTER
            site.scheduler.add_job(site.runCheck,
                                   trigger=trigType,
                                   kwargs={'name': name},
                                   id=name,
                                   **trigOpts)
    if not update:
        # refresh site map
        site.checkMap(True)
        _log.warning('Added check %s (%s) to site', name, newCheck.checkType)
    else:
        _log.warning('Updated check %s (%s)', name, newCheck.checkType)


def deleteAction(site, actionName):
    """Remove action from running site"""
    for checkName in site.checks:
        site.checks[checkName].del_action(actionName)
    del site.actions[actionName]
    _log.warning('Deleted action %s', actionName)


def updateUser(site, username, password=None, passHash=None):
    """Update or add user"""
    if site.webCfg is not None:
        if username:
            if password or passHash:
                if passHash is None:
                    passHash = createHash(password)
                site.webCfg['users'][username] = passHash
                _log.warning('Updated user %s', username)


def deleteUser(site, username):
    """Remove username from webui"""
    if site.webCfg is not None:
        if username and username != defaults.ADMINUSER:
            if username in site.webCfg['users']:
                del site.webCfg['users'][username]
                _log.warning('Deleted user %s', username)


def deleteCheck(site, checkName):
    """Remove check from running site"""
    # un-schedule
    job = site.scheduler.get_job(checkName)
    if job is not None:
        _log.debug('Removing %s (%r) from schedule', checkName, job)
        site.scheduler.remove_job(checkName)

    # remove
    if checkName in site.checks:
        tempCheck = site.checks[checkName]
        if tempCheck.remoteId is not None:
            del site.remotes[tempCheck.remoteId]
        del site.checks[checkName]

        # remove check from depends and sequences
        for name in site.checks:
            c = site.checks[name]
            c.del_depend(checkName)
            if c.checkType == 'sequence':
                c.del_check(checkName)

        topic = None
        if checkName in site.remoteTopics:
            topic = site.remoteTopics[checkName]
            _log.debug('Clear retain for remote check %s->%s', checkName,
                       topic)
            site.clearMsg(topic=topic)
            del site.remoteTopics[checkName]
        if tempCheck.publish:
            if tempCheck.publish != topic:
                _log.debug('Clear retain for check with publish %s->%s',
                           checkName, tempCheck.publish)
                site.clearMsg(topic=tempCheck.publish)

    # refresh site map
    site.checkMap(True)
    _log.warning('Deleted check %s from site', checkName)


def reorderSite(site, checkName, mode):
    """Reorder check in site as per mode, return True if changes made."""
    target = site.checks[checkName]
    siteMap = site.checkMap()
    if mode == 'dep':
        reMap = False
        if target.checkType != 'sequence':
            # Find first sequence this check belongs to and then check deps
            for seqName in siteMap:
                if checkName in siteMap[seqName]:
                    maxPri = -1
                    for dep in target.depends:
                        if dep in siteMap[seqName]:
                            maxPri = max(target.depends[dep].priority, maxPri)
                    if target.priority <= maxPri:
                        _log.debug('Adjusted priority on %s %d->%d', checkName,
                                   target.priority, maxPri + 1)
                        target.priority = maxPri + 1
                        reMap = True
                    break
        if reMap:
            siteMap = site.checkMap(True)

    if target.checkType == 'sequence':
        _log.debug('Re-order sequence check %r: %r', checkName, mode)
        if len(siteMap) < 3:
            _log.debug('Ignored request to reorder less than 2 sequences')
            return False
        else:
            seqOrd = [s for s in siteMap]
            seqOrd.remove(None)
            seqIdx = seqOrd.index(checkName)
            newIdx = seqIdx
            if mode == 'up':
                if seqIdx == 0:
                    _log.debug('Sequence already first')
                    return False
                newIdx -= 1
            elif mode == 'down':
                if seqIdx >= (len(seqOrd) - 1):
                    _log.debug('Sequence already last')
                    return False
                newIdx += 1
            elif mode == 'dep':
                pass

            # swap elements
            temp = seqOrd[seqIdx]
            seqOrd[seqIdx] = seqOrd[newIdx]
            seqOrd[newIdx] = temp

            # re-assign sequence priorities without changing subchecks
            priority = 100
            for seqName in seqOrd:
                site.checks[seqName].priority = priority
                priority += 100

            # force re-calculation of site map
            site.checkMap(True)
            return True
    else:
        _log.debug('Re-order check %r: %r', checkName, mode)
        # find first occurrence of check in map
        for seqName in siteMap:
            if checkName in siteMap[seqName]:
                seqStart = 10000
                if seqName is not None:
                    seqStart = site.checks[seqName].priority
                seqOrd = [c for c in siteMap[seqName]]
                if len(seqOrd) < 2:
                    _log.debug('Ignored request to reorder single check')
                    return False
                seqIdx = seqOrd.index(checkName)
                newIdx = seqIdx
                if mode == 'up':
                    if seqIdx == 0:
                        _log.debug('Check already first in sequence')
                        return False
                    newIdx -= 1
                elif mode == 'down':
                    if seqIdx >= (len(seqOrd) - 1):
                        _log.debug('Check already last in sequence')
                        return False
                    newIdx += 1
                elif mode == 'dep':
                    pass

                # swap elements
                temp = seqOrd[seqIdx]
                seqOrd[seqIdx] = seqOrd[newIdx]
                seqOrd[newIdx] = temp

                # re-assign sequence priorities without changing subchecks
                for checkName in seqOrd:
                    seqStart += 2  # Allow space for a dependent
                    site.checks[checkName].priority = seqStart

                # force re-calculation of site map
                site.checkMap(True)
                return True
    return False


def loadSite(site):
    """Load and initialise site"""
    cfg = None
    try:
        srcCfg = None
        with open(site.configFile) as f:
            srcCfg = json.load(f)

        if 'base' in srcCfg and isinstance(srcCfg['base'], str):
            site.base = srcCfg['base']

        if 'timezone' in srcCfg and isinstance(srcCfg['timezone'], str):
            site.timezone = getZone(srcCfg['timezone'])

        if 'webui' in srcCfg and isinstance(srcCfg['webui'], dict):
            site.webCfg = {}
            for k in defaults.WEBUICONFIG:
                if k in srcCfg['webui']:
                    site.webCfg[k] = srcCfg['webui'][k]
                else:
                    site.webCfg[k] = defaults.WEBUICONFIG[k]

        if 'mqtt' in srcCfg and isinstance(srcCfg['mqtt'], dict):
            site.mqttCfg = {}
            for k in defaults.MQTTCONFIG:
                if k in srcCfg['mqtt']:
                    site.mqttCfg[k] = srcCfg['mqtt'][k]
                else:
                    site.mqttCfg[k] = defaults.MQTTCONFIG[k]

        if 'log' in srcCfg and isinstance(srcCfg['log'], list):
            site.log = srcCfg['log']

        scheduler = AsyncIOScheduler()

        # load actions
        site.actions = {}
        if 'actions' in srcCfg and isinstance(srcCfg['actions'], dict):
            for a in srcCfg['actions']:
                nAct = loadAction(a, srcCfg['actions'][a])
                if nAct is not None:
                    site.actions[a] = nAct
                    _log.debug('Load action %r (%s)', a, nAct.actionType)

        # load checks
        site.checks = {}
        site.remotes = {}
        if 'checks' in srcCfg and isinstance(srcCfg['checks'], dict):
            for c in srcCfg['checks']:
                if isinstance(srcCfg['checks'][c], dict):
                    newCheck = loadCheck(c, srcCfg['checks'][c], site.timezone)
                    # add actions
                    if 'actions' in srcCfg['checks'][c]:
                        if isinstance(srcCfg['checks'][c]['actions'], list):
                            for a in srcCfg['checks'][c]['actions']:
                                if a in site.actions:
                                    newCheck.add_action(site.actions[a])
                                else:
                                    _log.info('%s ignored unknown action %s',
                                              c, a)
                    site.checks[c] = newCheck
                    if newCheck.remoteId is not None:
                        site.remotes[newCheck.remoteId] = c
                    _log.debug('Load check %r (%s)', c, newCheck.checkType)
        # patch the check dependencies, sequences and triggers
        for c in site.checks:
            if c in srcCfg['checks'] and 'depends' in srcCfg['checks'][c]:
                if isinstance(srcCfg['checks'][c]['depends'], list):
                    for d in srcCfg['checks'][c]['depends']:
                        if d in site.checks:
                            site.checks[c].add_depend(site.checks[d])
            if site.checks[c].checkType == 'sequence':
                if 'checks' in site.checks[c].options:
                    if isinstance(site.checks[c].options['checks'], list):
                        for s in site.checks[c].options['checks']:
                            if s in site.checks:
                                site.checks[c].add_check(site.checks[s])
            if site.checks[c].trigger is not None:
                trigOpts = None
                trigType = None
                if 'interval' in site.checks[c].trigger:
                    if isinstance(site.checks[c].trigger['interval'], dict):
                        trigOpts = dict(site.checks[c].trigger['interval'])
                        # set first trigger to run shortly after adding
                        trigOpts['start_date'] = datetime.now(UTC) + timedelta(
                            seconds=defaults.FIRSTTRIGGER)
                    trigType = 'interval'
                elif 'cron' in site.checks[c].trigger:
                    if isinstance(site.checks[c].trigger['cron'], dict):
                        trigOpts = dict(site.checks[c].trigger['cron'])
                    trigType = 'cron'
                if trigType is not None:
                    _log.debug('Adding %s trigger to check %s: %r', trigType,
                               c, trigOpts)
                    if 'timezone' not in trigOpts:
                        if site.checks[c].timezone:
                            trigOpts['timezone'] = site.checks[c].timezone
                        elif site.timezone:
                            trigOpts['timezone'] = site.timezone
                    trigOpts['jitter'] = defaults.TRIGGERJITTER
                    scheduler.add_job(site.runCheck,
                                      trigger=trigType,
                                      misfire_grace_time=None,
                                      kwargs={'name': c},
                                      id=c,
                                      **trigOpts)
                else:
                    _log.info('Invalid trigger for %s ignored', c)
                    site.checks[c].trigger = None

        site.scheduler = scheduler
        site.scheduler.start()
        # refresh site map
        site.checkMap(True)
    except Exception as e:
        _log.error('%s reading config: %s', e.__class__.__name__, e)


def mkCert(path, hostname):
    """Call openssl to make a self-signed certificate for hostname"""
    # Consider removal or replacement
    _log.debug('Creating self-signed SSL cert for %r at %r', hostname, path)
    crtTmp = None
    with NamedTemporaryFile(mode='w',
                            suffix='.tmp',
                            prefix='sav_',
                            dir=path,
                            delete=False) as f:
        crtTmp = f.name
    keyTmp = None
    with NamedTemporaryFile(mode='w',
                            suffix='.tmp',
                            prefix='sav_',
                            dir=path,
                            delete=False) as f:
        keyTmp = f.name
    crtOut = os.path.join(path, defaults.SSLCERT)
    keyOut = os.path.join(path, defaults.SSLKEY)
    template = """
[dn]
CN=%s
[req]
distinguished_name = dn
[EXT]
subjectAltName=DNS:%s
keyUsage=digitalSignature
extendedKeyUsage=serverAuth""" % (hostname, hostname)
    subject = '/CN=%s' % (hostname)
    cmd = [
        'openssl', 'req', '-x509', '-out', crtTmp, '-keyout', keyTmp,
        '-newkey', 'rsa:2048', '-nodes', '-sha256', '-subj', subject,
        '-extensions', 'EXT', '-config', '-'
    ]
    try:
        ret = run(cmd, input=template.encode('utf-8'), capture_output=True)
        if ret.returncode != 0:
            _log.error('Error creating SSL certificate: %s', ret.stderr)
        _log.debug('SSL certificate created OK')
        os.rename(crtTmp, crtOut)
        os.rename(keyTmp, keyOut)
    except Exception as e:
        _log.error('Error running openssl')



# SPDX-License-Identifier: MIT
"""Fletchck Web Interface"""

import asyncio
import ssl
import tornado.web
import tornado.ioloop
import tornado.template
import json
import os
from importlib.resources import files
from . import defaults
from . import util
from logging import getLogger, DEBUG, INFO, WARNING

_log = getLogger('fletchck.webui')
_log.setLevel(DEBUG)


class PackageLoader(tornado.template.BaseLoader):
    """Tornado template loader for importlib.files"""

    def resolve_path(self, name, parent_path=None):
        return name

    def _create_template(self, name):
        template = None
        ref = files('fletchck.templates').joinpath(name)
        if ref.is_file():
            with ref.open(mode='rb') as f:
                template = tornado.template.Template(f.read(),
                                                     name=name,
                                                     loader=self)
        else:
            _log.error('Unable to find named resource %s in templates', name)
        return template


class PackageFileHandler(tornado.web.StaticFileHandler):
    """Tornado static file handler for importlib.files"""

    @classmethod
    def get_absolute_path(cls, root, path):
        """Return the absolute path from importlib"""
        absolute_path = files('fletchck.static').joinpath(path)
        return absolute_path

    def validate_absolute_path(self, root, absolute_path):
        """Validate and return the absolute path"""
        if not absolute_path.is_file():
            raise tornado.web.HTTPError(404)
        return absolute_path

    @classmethod
    def get_content(cls, abspath, start=None, end=None):
        with abspath.open('rb') as file:
            if start is not None:
                file.seek(start)
            if end is not None:
                remaining = end - (start or 0)
            else:
                remaining = None
            while True:
                chunk_size = 64 * 1024
                if remaining is not None and remaining < chunk_size:
                    chunk_size = remaining
                chunk = file.read(chunk_size)
                if chunk:
                    if remaining is not None:
                        remaining -= len(chunk)
                    yield chunk
                else:
                    if remaining is not None:
                        assert remaining == 0
                    return

    def set_default_headers(self, *args, **kwargs):
        self.set_header("Content-Security-Policy", defaults.CSP)
        self.set_header("Strict-Transport-Security", "max-age=31536000")
        self.set_header("X-Frame-Options", "deny")
        self.set_header("X-Content-Type-Options", "nosniff")
        self.set_header("X-Permitted-Cross-Domain-Policies", "none")
        self.set_header("Referrer-Policy", "no-referrer")
        self.set_header("Cross-Origin-Embedder-Policy", "require-corp")
        self.set_header("Cross-Origin-Opener-Policy", "same-origin")
        self.set_header("Cross-Origin-Resource-Policy", "same-origin")
        self.clear_header("Server")


class Application(tornado.web.Application):

    def __init__(self, site):
        handlers = [
            (r"/", HomeHandler, dict(site=site)),
            (r"/checks", ChecksHandler, dict(site=site)),
            (r"/check/(.*)", CheckHandler, dict(site=site)),
            (r"/move/(.*)", MoveHandler, dict(site=site)),
            (r"/clone/(.*)", CloneHandler, dict(site=site)),
            (r"/config", ConfigHandler, dict(site=site)),
            (r"/action/(.*)", ActionHandler, dict(site=site)),
            (r"/user/(.*)", UserHandler, dict(site=site)),
            (r"/login", AuthLoginHandler, dict(site=site)),
            (r"/log", LogHandler, dict(site=site)),
            (r"/status", StatusHandler, dict(site=site)),
            (r"/logout", AuthLogoutHandler, dict(site=site)),
        ]
        templateLoader = PackageLoader(whitespace='oneline')
        settings = dict(
            site_version=defaults.VERSION,
            site_name=site.webCfg['name'],
            autoreload=False,
            serve_traceback=False,
            static_path='static',
            static_url_prefix='/s/',
            static_handler_class=PackageFileHandler,
            template_loader=templateLoader,
            cookie_secret=util.token_hex(32),
            login_url='/login?',
            debug=False,
        )
        super().__init__(handlers, **settings)


class BaseHandler(tornado.web.RequestHandler):

    def initialize(self, site):
        self._site = site

    def get_current_user(self):
        username = None
        cookie = self.get_signed_cookie("user",
                                        max_age_days=defaults.AUTHEXPIRY)
        if cookie is not None:
            username = cookie.decode('utf-8', 'replace')
            # invalidate sessions for a deleted user
            if username not in self._site.webCfg['users']:
                _log.warning('Rejected invalid session for deleted user: %s',
                             username)
                username = None
        return username

    def set_default_headers(self, *args, **kwargs):
        self.set_header("Content-Security-Policy", defaults.CSP)
        self.set_header("Strict-Transport-Security", "max-age=31536000")
        self.set_header("X-Frame-Options", "deny")
        self.set_header("X-Content-Type-Options", "nosniff")
        self.set_header("X-Permitted-Cross-Domain-Policies", "none")
        self.set_header("Referrer-Policy", "no-referrer")
        self.set_header("Cross-Origin-Embedder-Policy", "require-corp")
        self.set_header("Cross-Origin-Opener-Policy", "same-origin")
        self.set_header("Cross-Origin-Resource-Policy", "same-origin")
        self.clear_header("Server")


class HomeHandler(BaseHandler):
    """Site landing page."""

    @tornado.web.authenticated
    async def get(self):
        status = self._site.getStatus()
        self.render("home.html",
                    site=self._site,
                    status=status,
                    section='home')


class LogHandler(BaseHandler):
    """View of site logs."""

    @tornado.web.authenticated
    async def get(self):
        if self.get_argument('clear', ''):
            _log.info('Clearing volatile log')
            self._site.log.clear()
        status = self._site.getStatus()
        self.render("log.html", site=self._site, status=status, section='log')


class ChecksHandler(BaseHandler):
    """List of defined checks."""

    @tornado.web.authenticated
    async def get(self):
        status = self._site.getStatus()
        self.render("checks.html",
                    site=self._site,
                    status=status,
                    section='check')


class CloneHandler(BaseHandler):
    """Check cloning handler."""

    @tornado.web.authenticated
    async def get(self, path):
        check = None
        if path:
            if path in self._site.checks:
                status = self._site.getStatus()
                doSave = False
                checkName = path
                check = self._site.checks[checkName]
                newName = self._site.copyName(checkName)
                if check.checkType == 'remote':
                    _log.info('Unable to clone remote check %s', checkName)
                elif check.checkType == 'sequence':
                    _log.debug('Cloning sequence %s to %s', checkName, newName)
                    map = self._site.checkMap()
                    newConf = check.flatten()
                    del (newConf['data'])
                    oldChecks = []
                    if 'checks' in newConf['options']:
                        if newConf['options']['checks']:
                            oldChecks = newConf['options']['checks']
                            del (newConf['options']['checks'])

                    # determine and record clone sub check names
                    oldNew = {}
                    newChecks = []
                    for oldSubName in oldChecks:
                        if oldSubName in self._site.checks:
                            oldSubCheck = self._site.checks[oldSubName]
                            if oldSubCheck.checkType != 'remote':
                                newSubName = self._site.copyName(oldSubName)
                                oldNew[oldSubName] = newSubName
                                newChecks.append(newSubName)
                            else:
                                _log.debug(
                                    'Omitting remote check %s from clone',
                                    oldSubName)
                    newConf['options']['checks'] = newChecks

                    # create the cloned subchecks
                    for oldSubName in oldNew:
                        newSubName = oldNew[oldSubName]
                        _log.debug('Adding new subcheck: %s clone of %s',
                                   newSubName, oldSubName)
                        subCheck = self._site.checks[oldSubName]
                        newSubConf = subCheck.flatten()
                        del (newSubConf['data'])
                        newDeps = []
                        for depName in newSubConf['depends']:
                            if depName:
                                if depName in oldNew:
                                    newDeps.append(oldNew[depName])
                                else:
                                    newDeps.append(depName)
                        newSubConf['depends'] = newDeps
                        util.addCheck(self._site, newSubName, newSubConf)

                    # add sequence
                    util.addCheck(self._site, newName, newConf)
                    self._site.checkMap(True)
                    doSave = True
                else:
                    _log.debug('Cloning check %s to %s', checkName, newName)
                    newConf = check.flatten()
                    del (newConf['data'])
                    util.addCheck(self._site, newName, newConf)
                    if newName in self._site.checks:
                        newCheck = self._site.checks[newName]
                        if checkName in status['inseqs']:
                            for seqName in status['inseqs'][checkName]:
                                if seqName in self._site.checks:
                                    _log.debug(
                                        'Adding clone check %s to sequence %s',
                                        newName, seqName)
                                    seq = self._site.checks[seqName]
                                    seq.add_check(newCheck)
                            self._site.checkMap(True)
                    else:
                        _log.error('New check %s not yet on site', newName)
                    doSave = True
                if doSave:
                    await tornado.ioloop.IOLoop.current().run_in_executor(
                        None, self._site.saveConfig)

                self.redirect('/checks')
                return
            else:
                raise tornado.web.HTTPError(404)
        self.redirect('/checks')
        return


class MoveHandler(BaseHandler):
    """Check reordering handler."""

    @tornado.web.authenticated
    async def get(self, path):
        check = None
        if path:
            if path in self._site.checks:
                checkName = path
                mode = self.get_argument('m', 'down')
                if mode in ('down', 'up'):
                    if util.reorderSite(self._site, checkName, mode):
                        # save out config
                        await tornado.ioloop.IOLoop.current().run_in_executor(
                            None, self._site.saveConfig)
                else:
                    _log.warning('Ignored invalid mode mode')
                self.redirect('/checks')
                return
            else:
                raise tornado.web.HTTPError(404)
        self.redirect('/checks')
        return


class UserHandler(BaseHandler):
    """User editor."""

    @tornado.web.authenticated
    async def get(self, path):
        if self.current_user != defaults.ADMINUSER:
            _log.warning('User %s not authorised to edit users',
                         self.current_user)
            raise tornado.web.HTTPError(403)
        tmpPass = ''
        curUser = ''
        if path:
            if path in self._site.webCfg['users']:
                curUser = path
                if self.get_argument('delete', ''):
                    if curUser == defaults.ADMINUSER:
                        _log.warning('Administrator %s may not be deleted',
                                     defaults.ADMINUSER)
                        raise tornado.web.HTTPError(403)

                    _log.debug('Deleting user %s without undo', curUser)
                    self._site.deleteUser(curUser)
                    self.redirect('/config')
                    return
            else:
                raise tornado.web.HTTPError(404)
        else:
            # new user creation path
            tmpPass = util.randPass()
        status = self._site.getStatus()
        self.render("user.html",
                    status=status,
                    oldName=path,
                    user=curUser,
                    tmpPass=tmpPass,
                    section='config',
                    site=self._site,
                    adminuser=path == defaults.ADMINUSER,
                    formErrors=None)
        return

    @tornado.web.authenticated
    async def post(self, path):
        formErrors = []
        if self.current_user != defaults.ADMINUSER:
            _log.warning('User %s not authorised to edit users',
                         self.current_user)
            raise tornado.web.HTTPError(403)
        if path:
            if path in self._site.webCfg['users']:
                oldName = path
            else:
                raise tornado.web.HTTPError(404)
        oldName = self.get_argument('oldName', None)
        if oldName != path:
            _log.error('Form error: oldName does not match path request')
            raise tornado.web.HTTPError(500)
        newName = self.get_argument('username', '')
        if newName:
            if newName != oldName:
                if newName in self._site.webCfg['users']:
                    formErrors.append(
                        'Invalid username, please choose another')
            else:
                if oldName == defaults.ADMINUSER:
                    _log.error('Form error: Reject username change for admin')
                    raise tornado.web.HTTPError(500)
        else:
            if oldName == defaults.ADMINUSER:
                newName = oldName
            else:
                formErrors.append('Username required')

        newPass = self.get_argument('password', '')
        oldHash = None
        if not newPass:
            if path:
                oldHash = self._site.webCfg['users'][oldName]
            else:
                formErrors.append('Password required for new user')

        if formErrors:
            _log.info('User update %s with form errors', path)
            status = self._site.getStatus()
            self.render("user.html",
                        status=status,
                        oldName=path,
                        user=newName,
                        tmpPass=newPass,
                        section='config',
                        site=self._site,
                        adminuser=path == defaults.ADMINUSER,
                        formErrors=formErrors)
            return

        if newName != oldName:
            self._site.deleteUser(oldName)
        if newName != oldName or newPass:
            await tornado.ioloop.IOLoop.current().run_in_executor(
                None, self._site.updateUser, newName, newPass, oldHash)
            await tornado.ioloop.IOLoop.current().run_in_executor(
                None, self._site.saveConfig)
        else:
            _log.debug('No change required on user %s', oldName)
        self.redirect('/config')


class ActionHandler(BaseHandler):
    """Action editor."""

    @tornado.web.authenticated
    async def get(self, path):
        testMsg = None
        curAction = None
        if path:
            if path in self._site.actions:
                curAction = self._site.actions[path]
                runopt = self.get_argument('run', '')
                if self.get_argument('delete', ''):
                    _log.info('Deleting action %s without undo', path)
                    self._site.deleteAction(path)
                    self.redirect('/config')
                    return
                elif runopt:
                    _log.info('Manually running action %s', path)
                    res = await tornado.ioloop.IOLoop.current(
                    ).run_in_executor(None, self._site.testAction, path)
                    if res:
                        testMsg = 'Action test completed without error.'
                    else:
                        testMsg = 'Action test failed, check log for details.'

                    if runopt == 'list':
                        self.redirect('/config')
                        return
            else:
                raise tornado.web.HTTPError(404)
        else:
            curAction = util.loadAction(name='', config={'type': 'email'})
        status = self._site.getStatus()
        self.render("action.html",
                    status=status,
                    oldName=path,
                    curAction=curAction,
                    section='config',
                    site=self._site,
                    testMsg=testMsg,
                    formErrors=None)

    @tornado.web.authenticated
    async def post(self, path):
        oldConf = {}
        if path:
            if path in self._site.actions:
                oldConf = self._site.actions[path].flatten()
            else:
                raise tornado.web.HTTPError(404)

        # transfer form data into new config
        formErrors = []
        oldName = self.get_argument('oldName', None)
        if oldName != path:
            _log.error('Form error: oldName does not match path request')
            raise tornado.web.HTTPError(500)
        actionType = self.get_argument('actionType', None)
        actionName = self.get_argument('name', None)
        newConf = {"type": actionType}
        newConf['options'] = {}
        temp = self.get_argument('hostname', '')
        if temp:
            newConf['options']['hostname'] = util.toHostname(temp)
        for key in [
                'sender',
                'apikey',
                'url',
                'username',
                'password',
                'site',
        ]:
            temp = self.get_argument(key, '')
            if temp:
                newConf['options'][key] = temp
        recipients = self.get_argument('recipients', '')
        if recipients:
            newConf['options']['recipients'] = []
            recreg = set()
            for r in recipients.split():
                cr = r.lower()
                if cr not in recreg:
                    newConf['options']['recipients'].append(r)
                    recreg.add(cr)
        fallback = self.get_argument('fallback', '')
        if not fallback:
            if os.path.exists(defaults.SENDMAIL):
                fallback = defaults.SENDMAIL
        if fallback:
            newConf['options']['fallback'] = fallback

        for key in (
                'port',
                'timeout',
        ):
            temp = self.get_argument(key, '')
            if temp:
                newConf['options'][key] = int(temp)

        newInChecks = []
        for c in self.get_arguments('incheck'):
            if c in self._site.checks:
                newInChecks.append(c)

        if formErrors:
            _log.info('Action %s with form errors', path)
            status = self._site.getStatus()
            self.render("action.html",
                        status=status,
                        oldName=path,
                        curAction=curAction,
                        section='config',
                        site=self._site,
                        testMsg=testMsg,
                        formErrors=None)
            return

        # if form input ok - check changes
        async with self._site._lock:
            # remove old action from all checks
            for c in self._site.checks:
                self._site.checks[c].del_action(path)

            if path:
                # delete the old action handle
                self._site.deleteAction(path)

            # add action and add to any required checks
            self._site.addAction(actionName, newConf)
            curAction = self._site.actions[actionName]
            for c in newInChecks:
                if c in self._site.checks:
                    self._site.checks[c].add_action(curAction)

            # save out config
            await tornado.ioloop.IOLoop.current().run_in_executor(
                None, self._site.saveConfig)

        self.redirect('/action/' + self._site.pathQuote(actionName))


class CheckHandler(BaseHandler):
    """Check editor."""

    @tornado.web.authenticated
    async def get(self, path):
        check = None
        if path:
            if path in self._site.checks:
                check = self._site.checks[path]
                runopt = self.get_argument('run', '')
                if self.get_argument('delete', ''):
                    _log.info('Deleting %s without undo', path)
                    self._site.deleteCheck(path)
                    self.redirect('/checks')
                    return
                elif runopt:
                    _log.warning('Manually running %s', path)
                    await tornado.ioloop.IOLoop.current().run_in_executor(
                        None, self._site.runCheck, path)
                    retpath = '/check/' + self._site.pathQuote(path)
                    if runopt == 'list':
                        retpath = '/checks'
                    elif runopt == 'home':
                        retpath = '/'
                    self.redirect(retpath)
                    return
            else:
                raise tornado.web.HTTPError(404)
        else:
            check = util.loadCheck(name='',
                                   config={'type': 'ssh'},
                                   timezone=self._site.timezone)
            check.priority = 100 * len(self._site.checks)
        status = self._site.getStatus()
        self.render("check.html",
                    status=status,
                    oldName=path,
                    check=check,
                    section='check',
                    site=self._site,
                    formErrors=None)

    @tornado.web.authenticated
    async def post(self, path):
        oldConf = {}
        if path:
            if path in self._site.checks:
                oldConf = self._site.checks[path].flatten()
            else:
                raise tornado.web.HTTPError(404)

        # transfer form data into new config
        formErrors = []
        oldName = self.get_argument('oldName', None)
        if oldName != path:
            _log.error('Form error: oldName does not match path request')
            raise tornado.web.HTTPError(500)
        checkType = self.get_argument('checkType', None)
        checkName = self.get_argument('name', None)
        newConf = {"type": checkType}
        newConf['trigger'] = util.text2Trigger(self.get_argument(
            'trigger', ''))
        temp = self.get_argument('threshold', '')
        if temp:
            newConf['threshold'] = int(temp)
        temp = self.get_argument('retries', '')
        if temp:
            newConf['retries'] = int(temp)
        temp = self.get_argument('priority', '')
        if temp:
            newConf['priority'] = int(temp)
        else:
            if not path:
                # give new checks a default ordering > worst case: all seqs
                newConf['priority'] = 100 * len(self._site.checks)
        newConf['paused'] = bool(self.get_argument('paused', None))
        newConf['passAction'] = bool(self.get_argument('passAction', None))
        newConf['failAction'] = bool(self.get_argument('failAction', None))
        ptopic = self.get_argument('publish', None)
        if ptopic:
            newConf['publish'] = ptopic
        else:
            newConf['publish'] = None
        remid = self.get_argument('remoteId', None)
        if remid:
            newConf['remoteId'] = remid
        else:
            newConf['remoteId'] = None
        newConf['options'] = {}
        # hostname/ip
        temp = self.get_argument('hostname', '')
        if temp:
            newConf['options']['hostname'] = util.toHostname(temp)
        # string options
        for key in [
                'serialPort',
                'probe',
                'reqType',
                'reqPath',
                'reqName',
                'hostkey',
                'volume',
        ]:
            temp = self.get_argument(key, '')
            if temp:
                newConf['options'][key] = temp
        # timezone requires a little care
        temp = self.get_argument('timezone', '')
        if temp:
            newConf['options']['timezone'] = temp
            zinf = util.check.getZone(temp)
            if zinf is None:
                formErrors.append('Invalid timezone %r' % (temp))
        # int options
        for key in ['port', 'timeout', 'level', 'temperature', 'hysteresis']:
            temp = self.get_argument(key, '')
            if temp:
                newConf['options'][key] = int(temp)
        # tls & beeper default on
        temp = self.get_argument('tls', None)
        if not temp:
            newConf['options']['tls'] = False
        temp = self.get_argument('beeper', None)
        if not temp:
            newConf['options']['beeper'] = False
        # selfsigned/dnstcp is default off
        temp = self.get_argument('selfsigned', None)
        if temp:
            newConf['options']['selfsigned'] = True
        temp = self.get_argument('reqTcp', None)
        if temp:
            newConf['options']['reqTcp'] = True
        temp = self.get_arguments('checks')
        if temp:
            newConf['options']['checks'] = []
            for c in temp:
                if c:
                    newConf['options']['checks'].append(c)
        newConf['actions'] = self.get_arguments('actions')
        newConf['depends'] = self.get_arguments('depends')
        inseq = self.get_arguments('inseq')

        # final checks
        if not checkName:
            formErrors.append('Missing required check name')
        if not oldName or checkName != oldName:
            if checkName in self._site.checks:
                formErrors.append('Name already in use by another check')

        # build a temporary check object using the rest of the config
        check = util.loadCheck(name=checkName,
                               config=newConf,
                               timezone=self._site.timezone)
        for action in newConf['actions']:
            if action:
                if action in self._site.actions:
                    check.add_action(self._site.actions[action])
                else:
                    formErrors.append('Invalid action %r' % (action))
        dependReorder = False
        for depend in newConf['depends']:
            if depend:
                if depend in self._site.checks:
                    check.add_depend(self._site.checks[depend])
                    dependReorder = True
                else:
                    formErrors.append('Invalid check dependency %r' % (depend))
        inseqChange = False
        oldInseq = []
        status = self._site.getStatus()
        if oldName in status['inseqs']:
            oldInseq = status['inseqs'][oldName]
        newInseq = []
        for seq in inseq:
            if seq:
                if seq in self._site.checks:
                    if self._site.checks[seq].checkType == 'sequence':
                        newInseq.append(seq)
                    else:
                        formErrors.append('Not a sequence %r' % (seq, ))
                else:
                    formErrors.append('Invalid sequence %r' % (seq, ))
        if newInseq != oldInseq:
            inseqChange = True
            _log.info('Sequence membership change: %r != %r', newInseq,
                      oldInseq)

        if formErrors:
            _log.info('Edit check %s with form errors', path)
            status = self._site.getStatus()
            self.render("check.html",
                        status=status,
                        oldName=path,
                        check=check,
                        site=self._site,
                        section='check',
                        formErrors=formErrors)
            return

        if 'data' in oldConf:
            newConf['data'] = oldConf['data']

        # patch remoteId if name changes on a remote check
        if check.checkType == 'remote':
            if oldName and checkName != oldName and check.remoteId is None:
                _log.debug('Using oldname=%r for remoteId on remote check %r',
                           oldName, checkName)
                newConf['remoteId'] = oldName

        # if form input ok - check changes
        async with self._site._lock:
            runCheck = False
            if path:
                _log.info('Saving changes to check %s', path)
                self._site.updateCheck(path, checkName, newConf)
            else:
                _log.info('Saving new check %s', checkName)
                self._site.addCheck(checkName, newConf)
            if dependReorder:
                util.reorderSite(self._site, checkName, 'dep')
            if inseqChange:
                for seqName in oldInseq:
                    if seqName not in newInseq:
                        _log.info('Remove %s from sequence %s', checkName,
                                  seqName)
                        self._site.checks[seqName].del_check(checkName)
                if checkName in self._site.checks:
                    newCheck = self._site.checks[checkName]
                    for seqName in newInseq:
                        if seqName not in oldInseq:
                            _log.info('Add %s to sequence %s', checkName,
                                      seqName)
                            self._site.checks[seqName].add_check(newCheck)
                else:
                    raise RuntimeError('Check object/name mismatch %s' %
                                       (checkName, ))
                # force re-read and re-order of map
                self._site.checkMap(True)
                dependReorder = True
            if dependReorder:
                util.reorderSite(self._site, checkName, 'dep')

            # save out config
            await tornado.ioloop.IOLoop.current().run_in_executor(
                None, self._site.saveConfig)

        # run check and wait for result
        if runCheck:
            _log.info('Running initial test on %s', checkName)
            await tornado.ioloop.IOLoop.current().run_in_executor(
                None, self._site.runCheck, checkName)

        self.redirect('/check/' + self._site.pathQuote(checkName))


class ConfigHandler(BaseHandler):

    @tornado.web.authenticated
    async def get(self):
        status = self._site.getStatus()
        self.render("config.html",
                    status=status,
                    section='config',
                    site=self._site,
                    showusers=self.current_user == defaults.ADMINUSER,
                    mqttCfg=self._site.mqttCfg,
                    webCfg=self._site.webCfg,
                    formErrors=[])

    @tornado.web.authenticated
    async def post(self):
        formErrors = []
        tzUpdate = False
        temp = self.get_argument('timezone', '')
        if temp:
            zinf = util.check.getZone(temp)
            if zinf is not None:
                if self._site.timezone != zinf:
                    tzUpdate = True
                    self._site.timezone = zinf
            else:
                formErrors.append('Invalid timezone %r' % (temp))
        else:
            if self._site.timezone is not None:
                tzUpdate = True
                self._site.timezone = None

        newCfg = {'mqtt': None, 'webui': {}}
        for mkey in defaults.WEBUICONFIG:
            if mkey != 'users':
                newCfg['webui'][mkey] = defaults.WEBUICONFIG[mkey]
                if mkey in ('hostname', 'cert', 'key', 'name'):
                    temp = self.get_argument('webui.' + mkey, None)
                    if temp:
                        newCfg['webui'][mkey] = temp
                elif mkey == 'port':
                    temp = self.get_argument('webui.port', '')
                    if temp:
                        nv = None
                        if temp.isdigit():
                            temp = int(temp)
                            if temp > 0 and temp < 65536:
                                nv = temp
                        if nv is not None:
                            newCfg['webui']['port'] = nv
                        else:
                            formErrors.append('Invalid web UI port %r' %
                                              (temp))
                    else:
                        formErrors.append('Missing required Web UI port')

        enableMqtt = bool(self.get_argument('mqtt.enable', None))
        if enableMqtt:
            newCfg['mqtt'] = {}
            for mkey in defaults.MQTTCONFIG:
                newCfg['mqtt'][mkey] = defaults.MQTTCONFIG[mkey]
                if mkey in ('tls', 'persist', 'retain', 'autoadd'):
                    newCfg['mqtt'][mkey] = bool(
                        self.get_argument('mqtt.' + mkey, None))
                elif mkey in ('hostname', 'username', 'password', 'clientid',
                              'basetopic'):
                    temp = self.get_argument('mqtt.' + mkey, None)
                    if temp:
                        newCfg['mqtt'][mkey] = temp
                elif mkey == 'port':
                    temp = self.get_argument('mqtt.port', '')
                    if temp:
                        nv = None
                        if temp.isdigit():
                            temp = int(temp)
                            if temp > 0 and temp < 65536:
                                nv = temp
                        if nv is not None:
                            newCfg['mqtt']['port'] = nv
                        else:
                            formErrors.append('Invalid MQTT port %r' % (temp))
                elif mkey == 'qos':
                    temp = self.get_argument('mqtt.qos', '')
                    if temp:
                        nv = None
                        if temp.isdigit():
                            temp = int(temp)
                            if temp in (0, 1, 2):
                                nv = temp
                        if nv is not None:
                            newCfg['mqtt']['qos'] = nv
                        else:
                            formErrors.append('Invalid MQTT QoS %r' % (temp))

        if formErrors:
            _log.info('Edit config with form errors')
            status = self._site.getStatus()
            self.render("config.html",
                        status=status,
                        section='config',
                        site=self._site,
                        showusers=self.current_user == defaults.ADMINUSER,
                        mqttCfg=newCfg['mqtt'],
                        webCfg=newCfg['webui'],
                        formErrors=formErrors)
            return

        mqttUpdate = False
        if newCfg['mqtt'] is None:
            if self._site.mqttCfg is not None:
                mqttUpdate = True
                _log.debug('Disabling MQTT interface on site')
            self._site.mqttCfg = None
            self._site.reconnectMqtt()
        else:
            if self._site.mqttCfg is None:
                self._site.mqttCfg = {}
                for mkey in defaults.MQTTCONFIG:
                    self._site.mqttCfg[mkey] = defaults.MQTTCONFIG[mkey]
            for mkey in defaults.MQTTCONFIG:
                if mkey != 'debug':
                    if self._site.mqttCfg[mkey] != newCfg['mqtt'][mkey]:
                        mqttUpdate = True
                        self._site.mqttCfg[mkey] = newCfg['mqtt'][mkey]
            if mqttUpdate:
                _log.debug('Updating MQTT interface on site')
                self._site.reconnectMqtt()

        webuiUpdate = False
        for mkey in defaults.WEBUICONFIG:
            if mkey != 'users':
                if self._site.webCfg[mkey] != newCfg['webui'][mkey]:
                    webuiUpdate = True
                    self._site.webCfg[mkey] = newCfg['webui'][mkey]

        if tzUpdate or mqttUpdate or webuiUpdate:
            async with self._site._lock:
                await tornado.ioloop.IOLoop.current().run_in_executor(
                    None, self._site.saveConfig)
        if webuiUpdate:
            newUrl = 'https://%s:%s' % (self._site.webCfg['hostname'],
                                        self._site.webCfg['port'])
            # TODO: add a deferred task to restart webui (somehow)
            _log.warning('Web UI restarting at %s', newUrl)
            status = self._site.getStatus()
            self.render("reload.html",
                        status=status,
                        section='config',
                        site=self._site,
                        url=newUrl,
                        message='Web UI restarting...')
            self._site.runWebRestart()
            return
        else:
            self.redirect('/config')


class StatusHandler(BaseHandler):

    @tornado.web.authenticated
    async def get(self):
        status = self._site.getStatus()
        self.set_header("Content-Type", 'application/json')
        self.write(json.dumps(status))


class AuthLoginHandler(BaseHandler):

    async def get(self):
        self.render("login.html", error=None)

    async def post(self):
        await asyncio.sleep(0.3 + util.randbits(10) / 3000)
        un = self.get_argument('username', '')
        pw = self.get_argument('password', '')
        hash = None
        uv = None
        if un and un in self._site.webCfg['users']:
            hash = self._site.webCfg['users'][un]
            uv = un
        else:
            hash = self._site.webCfg['users']['']
            uv = None

        # checkPass has a long execution by design
        po = await tornado.ioloop.IOLoop.current().run_in_executor(
            None, util.checkPass, pw, hash)

        if uv is not None and po:
            self.set_signed_cookie("user",
                                   uv.encode('utf-8', 'replace'),
                                   expires_days=None,
                                   secure=True,
                                   samesite='Strict')
            _log.warning('Login username=%r (%s)', un, self.request.remote_ip)
            self.redirect('/')
        else:
            _log.warning('Invalid login username=%r (%s)', un,
                         self.request.remote_ip)
            self.render("login.html", error='Invalid login details')


class AuthLogoutHandler(BaseHandler):

    def get(self):
        self.clear_cookie("user", secure=True, samesite='Strict')
        self.set_header("Clear-Site-Data", '"*"')
        self.redirect('/login')


def loadUi(site):
    app = Application(site)
    ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
    ssl_ctx.load_cert_chain(site.webCfg['cert'], site.webCfg['key'])
    port = site.webCfg['port']
    if site.webUiPort is not None:
        port = site.webUiPort
    srv = tornado.httpserver.HTTPServer(app, ssl_options=ssl_ctx)
    srv.listen(port, address=site.webCfg['hostname'])
    _log.info('Web UI listening on: https://%s:%s', site.webCfg['hostname'],
              port)
    return srv



# SPDX-License-Identifier: MIT
"""Machine check classes"""

from datetime import datetime, timezone
# Python < 3.11 workaround for datetime.UTC
UTC = timezone.utc
from zoneinfo import ZoneInfo
from dateutil.parser import parse as dateparse
from . import defaults
from logging import getLogger, DEBUG, INFO, WARNING, ERROR
from smtplib import SMTP, SMTP_SSL
from imaplib import IMAP4_SSL, IMAP4_SSL_PORT
from http.client import HTTPSConnection, HTTPConnection
from paramiko.transport import Transport as SSH
from threading import Lock
from cryptography import x509
from .ups import UpsQsV
from shutil import disk_usage
from psutil import virtual_memory, cpu_percent
from time import sleep
import dns.rdatatype
import dns.name
import dns.resolver
import dns.reversename
import ssl
import socket
try:
    from lxml import etree
except ImportError:
    import xml.etree.ElementTree as etree

_log = getLogger('fletchck.check')
_log.setLevel(DEBUG)
getLogger('paramiko.transport').setLevel(INFO)

CHECK_TYPES = {}
_TERA = 1024 * 1024 * 1024 * 1024
_GIGA = 1024 * 1024 * 1024

# Serial port locks
_serialLock = {'': Lock()}

# Temporary: Common local timezone labels
LOCALZONES = {
    "AEST": +36000,
    "AEDT": +39600,
    "ACST": +34200,
    "ACDT": +37800,
}


def timeString(timezone=None):
    return datetime.now().astimezone(timezone).strftime("%d %b %Y %H:%M %Z")


def getZone(timezone=None):
    """Return a zoneinfo if possible"""
    ret = None
    try:
        ret = ZoneInfo(timezone)
    except Exception:
        _log.warning('Ignored invalid timezone %r', timezone)
    return ret


def certExpiry(cert):
    """Raise SSL certificate error if about to expire"""
    if cert is not None and 'notAfter' in cert:
        expiry = ssl.cert_time_to_seconds(cert['notAfter'])
        nowsecs = datetime.now(UTC).timestamp()
        daysLeft = (expiry - nowsecs) // 86400
        _log.debug('Certificate %r expiry %r: %d days', cert['subject'],
                   cert['notAfter'], daysLeft)
        if daysLeft < defaults.CERTEXPIRYDAYS:
            raise ssl.SSLCertVerificationError(
                'Certificate expires in %d days' % (daysLeft))
    else:
        _log.debug('Certificate missing - expiry check skipped')
    return True


def loadCheck(name, config, timezone=None):
    """Create and return a check object for the provided flat config"""
    ret = None
    if config['type'] in CHECK_TYPES:
        options = defaults.getOpt('options', config, dict, {})
        ret = CHECK_TYPES[config['type']](name, options)
        ret.checkType = config['type']
        ret.timezone = timezone
        if 'trigger' in config and isinstance(config['trigger'], dict):
            ret.trigger = config['trigger']
        if 'threshold' in config and isinstance(config['threshold'], int):
            if config['threshold'] > 0:
                ret.threshold = config['threshold']
        if 'retries' in config and isinstance(config['retries'], int):
            if config['retries'] > 0:
                ret.retries = config['retries']
        if 'subType' in config and isinstance(config['subType'], str):
            ret.subType = config['subType']
        if 'priority' in config and isinstance(config['priority'], int):
            ret.priority = config['priority']
        if 'paused' in config and isinstance(config['paused'], bool):
            ret.paused = config['paused']
        if 'failAction' in config and isinstance(config['failAction'], bool):
            ret.failAction = config['failAction']
        if 'passAction' in config and isinstance(config['passAction'], bool):
            ret.passAction = config['passAction']
        if 'publish' in config and isinstance(config['publish'], str):
            ret.publish = config['publish']
        if 'remoteId' in config and isinstance(config['remoteId'], str):
            ret.remoteId = config['remoteId']
        if 'timezone' in options and isinstance(options['timezone'], str):
            ret.timezone = getZone(options['timezone'])
        if 'data' in config:
            if 'failState' in config['data']:
                if isinstance(config['data']['failState'], (bool, str)):
                    ret.failState = config['data']['failState']
            if 'failCount' in config['data']:
                if isinstance(config['data']['failCount'], int):
                    if config['data']['failCount'] >= 0:
                        ret.failCount = config['data']['failCount']
            if 'threshold' in config['data']:
                if isinstance(config['data']['threshold'], int):
                    if config['data']['threshold'] >= 0:
                        ret.threshold = config['data']['threshold']
            if 'lastFail' in config['data']:
                if isinstance(config['data']['lastFail'], str):
                    ret.lastFail = config['data']['lastFail']
            if 'lastPass' in config['data']:
                if isinstance(config['data']['lastPass'], str):
                    ret.lastPass = config['data']['lastPass']
            if 'lastCheck' in config['data']:
                if isinstance(config['data']['lastCheck'], str):
                    ret.lastCheck = config['data']['lastCheck']
            if 'lastUpdate' in config['data']:
                if isinstance(config['data']['lastUpdate'], str):
                    ret.lastUpdate = config['data']['lastUpdate']
            if 'softFail' in config['data']:
                if isinstance(config['data']['softFail'], str):
                    ret.softFail = config['data']['softFail']
            if 'level' in config['data']:
                if isinstance(config['data']['level'], str):
                    ret.level = config['data']['level']
            if 'log' in config['data']:
                if isinstance(config['data']['log'], list):
                    ret.log = config['data']['log']
    else:
        _log.warning('Invalid check type ignored')
    return ret


class BaseCheck():
    """Check base class"""

    def __init__(self, name, options={}):
        self.name = name
        self.paused = False
        self.failAction = True
        self.passAction = True
        self.publish = None
        self.remoteId = None
        self.threshold = 1
        self.retries = 1
        self.priority = 0
        self.options = options
        self.checkType = None
        self.subType = None
        self.trigger = None
        self.timezone = None
        self.level = None

        self.actions = {}
        self.depends = {}

        self.failState = True
        self.softFail = None
        self.failCount = 0
        self.log = []
        self.oldLog = None
        self.lastFail = None
        self.lastPass = None
        self.lastCheck = None
        self.lastUpdate = None

    def _runCheck(self):
        """Perform the required check and return fail state"""
        return False

    def timeString(self, dt):
        """Return string formatted datetime dt in check's timezone"""
        return dt.astimezone(self.timezone).strftime("%d %b %Y %H:%M %Z")

    def getState(self):
        """Return a string indicating pass or fail"""
        if self.paused:
            return 'PAUSED'
        elif self.failState:
            return 'FAIL'
        else:
            return 'PASS'

    def getSummary(self):
        """Return a short text summary of the check state"""
        ret = ''
        if self.failState and self.log:
            ret = self.log[-1]
        return ret

    def notify(self):
        """Trigger all configured actions"""
        for action in self.actions:
            self.actions[action].trigger(self)

    def update(self):
        """Run check, update state and trigger events as required"""
        if self.paused:
            _log.debug('%s (%s) PAUSED', self.name, self.checkType)
            self.log = ['PAUSED']
            self.failState = False
            return False
        thisTime = timeString(self.timezone)
        self.lastCheck = thisTime
        self.softFail = None
        for d in self.depends:
            if self.depends[d].failState:
                self.softFail = d
                _log.info('%s (%s) SOFTFAIL (depends=%s) %s', self.name,
                          self.checkType, d, thisTime)
                self.log = ['SOFTFAIL (depends=%s)' % (d)]
                return True

        self.oldLog = self.log
        count = 0
        while count < self.retries and self.softFail is None:
            self.log = []
            count += 1
            if count > 1:
                self.log.append('Retry %d/%d' % (count, self.retries))
                _log.info('%s (%s): Retrying %d/%d', self.name, self.checkType,
                          count, self.retries)
            curFail = self._runCheck()
            if not curFail:
                break
        if count > 1 and curFail:
            self.log.append('Failed after %d tries' % (count, ))

        _log.info(
            '%s (%s): %s curFail=%r prevFail=%r failCount=%r level=%r %s',
            self.name, self.checkType, self.getState(), curFail,
            self.failState, self.failCount, self.level, thisTime)

        if curFail:
            self.failCount += 1
            if self.failCount >= self.threshold:
                # compare fail state by value
                if curFail != self.failState:
                    _log.warning('%s (%s) Log: %r', self.name, self.checkType,
                                 self.log)
                    _log.warning('%s (%s) FAIL', self.name, self.checkType)
                    self.failState = curFail
                    self.lastFail = thisTime
                    if self.failAction:
                        self.notify()
        else:
            self.failCount = 0
            if self.failState:
                _log.warning('%s (%s) PASS', self.name, self.checkType)
                self.failState = curFail
                self.lastPass = thisTime
                if self.passAction:
                    self.notify()

        return self.failState

    def add_action(self, action):
        """Add the specified action"""
        self.actions[action.name] = action

    def del_action(self, name):
        """Remove the specified action"""
        if name in self.actions:
            del self.actions[name]

    def add_depend(self, check):
        """Add check to the set of dependencies"""
        if check is not self:
            self.depends[check.name] = check
            _log.debug('Added dependency %s to %s', check.name, self.name)

    def del_depend(self, name):
        """Remove check from the set of dependencies"""
        if name in self.depends:
            del self.depends[name]
            _log.debug('Removed dependency %s from %s', name, self.name)

    def replace_depend(self, name, check):
        """Replace dependency with new entry if it existed"""
        if name in self.depends:
            self.del_depend(name)
            self.add_depend(check)

    def getStrOpt(self, key, default=None):
        return defaults.getOpt(key, self.options, str, default)

    def getBoolOpt(self, key, default=None):
        return defaults.getOpt(key, self.options, bool, default)

    def getIntOpt(self, key, default=None):
        return defaults.getOpt(key, self.options, int, default)

    def msgObj(self):
        """Return a remote notification object for this check"""
        return {
            'name': self.name,
            'type': self.checkType,
            'data': {
                'threshold': self.threshold,
                'failState': self.failState,
                'failCount': self.failCount,
                'log': self.log,
                'softFail': self.softFail,
                'lastCheck': self.lastCheck,
                'lastFail': self.lastFail,
                'lastPass': self.lastPass,
                'level': self.level
            }
        }

    def flatten(self):
        """Return the check as a flattened dictionary"""
        actList = [a for a in self.actions]
        depList = [d for d in self.depends]
        optMap = {}
        for key in self.options:
            optMap[key] = self.options[key]
        return {
            'type': self.checkType,
            'subType': self.subType,
            'trigger': self.trigger,
            'threshold': self.threshold,
            'retries': self.retries,
            'priority': self.priority,
            'paused': self.paused,
            'failAction': self.failAction,
            'passAction': self.passAction,
            'publish': self.publish,
            'remoteId': self.remoteId,
            'options': optMap,
            'actions': actList,
            'depends': depList,
            'data': {
                'failState': self.failState,
                'failCount': self.failCount,
                'log': self.log,
                'softFail': self.softFail,
                'lastCheck': self.lastCheck,
                'lastUpdate': self.lastUpdate,
                'lastFail': self.lastFail,
                'lastPass': self.lastPass,
                'level': self.level
            }
        }


class submitCheck(BaseCheck):
    """SMTP-over-SSL / submissions check"""

    def _runCheck(self):
        hostname = self.getStrOpt('hostname', '')
        port = self.getIntOpt('port', 0)
        timeout = self.getIntOpt('timeout', defaults.SUBMITTIMEOUT)
        selfsigned = self.getBoolOpt('selfsigned', False)

        failState = True
        try:
            ctx = ssl.create_default_context()
            if selfsigned:
                ctx.check_hostname = False
                ctx.verify_mode = ssl.CERT_NONE
            with SMTP_SSL(host=hostname,
                          port=port,
                          timeout=timeout,
                          context=ctx) as s:
                self.log.append(repr(s.ehlo()))
                failState = False
                self.log.append(repr(s.noop()))
                self.log.append(repr(s.quit()))
        except Exception as e:
            if isinstance(e, ssl.SSLCertVerificationError):
                self.softFail = 'certificate'
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       hostname, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' % (hostname, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType, hostname,
                   failState)
        return failState


class smtpCheck(BaseCheck):
    """SMTP service check"""

    def _runCheck(self):
        tls = self.getBoolOpt('tls', True)
        hostname = self.getStrOpt('hostname', '')
        port = self.getIntOpt('port', 0)
        timeout = self.getIntOpt('timeout', defaults.SMTPTIMEOUT)
        selfsigned = self.getBoolOpt('selfsigned', False)

        failState = True
        try:
            with SMTP(host=hostname, port=port, timeout=timeout) as s:
                if tls:
                    ctx = ssl.create_default_context()
                    if selfsigned:
                        ctx.check_hostname = False
                        ctx.verify_mode = ssl.CERT_NONE
                    self.log.append(repr(s.starttls(context=ctx)))
                    certExpiry(s.sock.getpeercert())
                self.log.append(repr(s.ehlo()))
                failState = False
                self.log.append(repr(s.noop()))
                self.log.append(repr(s.quit()))
        except Exception as e:
            if isinstance(e, ssl.SSLCertVerificationError):
                self.softFail = 'certificate'
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       hostname, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' % (hostname, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType, hostname,
                   failState)
        return failState


class imapCheck(BaseCheck):
    """IMAP4+SSL service check"""

    def _runCheck(self):
        hostname = self.getStrOpt('hostname', '')
        port = self.getIntOpt('port', IMAP4_SSL_PORT)
        timeout = self.getIntOpt('timeout', defaults.IMAPTIMEOUT)
        selfsigned = self.getBoolOpt('selfsigned', False)

        failState = True
        try:
            ctx = ssl.create_default_context()
            if selfsigned:
                ctx.check_hostname = False
                ctx.verify_mode = ssl.CERT_NONE
            with IMAP4_SSL(host=hostname,
                           port=port,
                           ssl_context=ctx,
                           timeout=timeout) as i:
                certExpiry(i.sock.getpeercert())
                self.log.append(repr(i.noop()))
                self.log.append(repr(i.logout()))
                failState = False
        except Exception as e:
            if isinstance(e, ssl.SSLCertVerificationError):
                self.softFail = 'certificate'
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       hostname, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' % (hostname, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType, hostname,
                   failState)
        return failState


class certCheck(BaseCheck):
    """TLS Certificate check"""

    def _runCheck(self):
        hostname = self.getStrOpt('hostname', '')
        port = self.getIntOpt('port')
        timeout = self.getIntOpt('timeout', defaults.CERTTIMEOUT)
        selfsigned = self.getBoolOpt('selfsigned', False)
        probe = self.getStrOpt('probe')
        self.level = None

        failState = True
        try:
            if not selfsigned:
                # do full TLS negotiation
                ctx = ssl.create_default_context()
                conn = ctx.wrap_socket(socket.create_connection(
                    (hostname, port), timeout=timeout),
                                       server_hostname=hostname)
                certExpiry(conn.getpeercert())
                if probe is not None:
                    self.log.append(
                        'send: %r, %r' %
                        (probe, conn.sendall(probe.encode('utf-8'))))
                    self.log.append('recv: %r' % (conn.recv(1024)))
                conn.shutdown(socket.SHUT_RDWR)
                conn.close()
            else:
                # check expiry
                pemCert = ssl.get_server_certificate(addr=(hostname, port))
                # timeout=timeout)  # timeout added in python3.10
                cert = x509.load_pem_x509_certificate(pemCert.encode('ascii'))
                expiry = cert.not_valid_after_utc
                remain = expiry - datetime.now(UTC)
                daysLeft = int(remain.total_seconds() // 86400)
                self.level = '%d days' % (daysLeft, )
                _log.debug('Certificate %r expiry %r: %d days', hostname,
                           expiry.astimezone().isoformat(), daysLeft)
                if daysLeft < defaults.CERTEXPIRYDAYS:
                    raise ssl.SSLCertVerificationError(
                        'Certificate expires in %d days' % (daysLeft))
                _log.debug('%s (%s) %s: Certificate not verified', self.name,
                           self.checkType, hostname)
            failState = False
        except Exception as e:
            if isinstance(e, ssl.SSLCertVerificationError):
                self.softFail = 'certificate'
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       hostname, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' % (hostname, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType, hostname,
                   failState)
        return failState


class dnsCheck(BaseCheck):
    """DNS service check"""

    def _runCheck(self):
        hostname = self.getStrOpt('hostname', '127.0.0.53')
        timeout = self.getIntOpt('timeout', defaults.DNSTIMEOUT)
        port = self.getIntOpt('port', defaults.DNSPORT)
        reqName = self.getStrOpt('reqName', 'org')
        reqType = self.getStrOpt('reqType', 'soa')
        reqTcp = self.getBoolOpt('reqTcp', defaults.DNSTCP)
        failState = True

        try:
            reqType = dns.rdatatype.from_text(reqType)
            if reqType is dns.rdatatype.PTR:
                reqName = dns.reversename.from_address(reqName)
            else:
                reqName = dns.name.from_unicode(reqName)

            a = dns.resolver.resolve_at(where=hostname,
                                        qname=reqName,
                                        rdtype=reqType,
                                        port=port,
                                        tcp=reqTcp,
                                        lifetime=timeout)

            self.log.append(
                repr((str(reqName), reqType.name, ' '.join(
                    (str(r) for r in a)))))
            failState = False
        except Exception as e:
            if isinstance(e, ssl.SSLCertVerificationError):
                self.softFail = 'certificate'
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       hostname, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' % (hostname, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType, hostname,
                   failState)
        return failState


class httpsCheck(BaseCheck):
    """HTTPS service check"""

    def _runCheck(self):
        tls = self.getBoolOpt('tls', True)
        hostname = self.getStrOpt('hostname', '')
        port = self.getIntOpt('port')
        timeout = self.getIntOpt('timeout', defaults.HTTPSTIMEOUT)
        selfsigned = self.getBoolOpt('selfsigned', False)
        reqType = self.getStrOpt('reqType', 'HEAD')
        reqPath = self.getStrOpt('reqPath', '/')

        failState = True
        try:
            h = None
            if tls:
                ctx = ssl.create_default_context()
                if selfsigned:
                    ctx.check_hostname = False
                    ctx.verify_mode = ssl.CERT_NONE
                h = HTTPSConnection(host=hostname,
                                    port=port,
                                    timeout=timeout,
                                    context=ctx)
                h.request(reqType, reqPath)
                certExpiry(h.sock.getpeercert())
            else:
                _log.debug('Plain http check requested')
                h = HTTPConnection(host=hostname, port=port, timeout=timeout)
                h.request(reqType, reqPath)
            r = h.getresponse()
            self.log.append(repr((r.status, r.headers.as_string())))
            failState = False
        except Exception as e:
            if isinstance(e, ssl.SSLCertVerificationError):
                self.softFail = 'certificate'
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       hostname, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' % (hostname, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType, hostname,
                   failState)
        return failState


class sshCheck(BaseCheck):
    """SSH service check"""

    def _runCheck(self):
        hostname = self.getStrOpt('hostname', '')
        port = self.getIntOpt('port', 22)
        timeout = self.getIntOpt('timeout', defaults.SSHTIMEOUT)
        hostkey = self.getStrOpt('hostkey')

        failState = True
        try:
            t = SSH(socket.create_connection((hostname, port),
                                             timeout=timeout))
            t.start_client(timeout=timeout)
            hk = t.get_remote_server_key().get_base64()
            self.log.append('%s:%d %r' % (hostname, port, hk))
            if hostkey is not None and hostkey != hk:
                raise ValueError('Invalid host key')
            elif hostkey is None:
                _log.info('%s (%s) %s: Adding hostkey=%s', self.name,
                          self.checkType, hostname, hk)
                self.options['hostkey'] = hk
            self.log.append('ignore: %r' % (t.send_ignore()))
            self.log.append('close: %r' % (t.close()))
            failState = False
        except Exception as e:
            if isinstance(e, ValueError):
                self.softFail = 'hostkey'
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       hostname, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' % (hostname, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType, hostname,
                   failState)
        return failState


class upsStatus(BaseCheck):
    """UPS basic check"""

    def _runCheck(self):
        serialPort = self.getStrOpt('serialPort', '')
        if serialPort:
            if serialPort not in _serialLock:
                with _serialLock['']:
                    _serialLock[serialPort] = Lock()
        beeper = self.getBoolOpt('beeper', True)

        failState = True
        try:
            _log.debug('Waiting for serialport')
            with _serialLock[serialPort]:
                u = UpsQsV(serialPort)
                u.setBeeper(beeper)
                self.log.append('Load: %d%%, Battery: %0.1fV' %
                                (u.load, u.battery))
                self.log.append(u.getInfo(update=False))
                if u.lowBattery:
                    self.log.append('Low battery warning: %0.1fV' %
                                    (u.battery))
                failState = (u.error or u.fail or u.fault or u.lowBattery
                             or u.shutdown)
        except Exception as e:
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       serialPort, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' %
                            (serialPort, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType,
                   serialPort, failState)
        return failState


class upsTest(BaseCheck):
    """Run a UPS self-test and check result"""

    def _runCheck(self):
        serialPort = self.getStrOpt('serialPort', '')
        if serialPort:
            if serialPort not in _serialLock:
                with _serialLock['']:
                    _serialLock[serialPort] = Lock()

        failState = True
        try:
            _log.debug('Waiting for serialport')
            with _serialLock[serialPort]:
                u = UpsQsV(serialPort)
                failState, msg = u.runTest()
                self.log.append(msg)
                _log.info('%s (%s) %s: %s', self.name, self.checkType,
                          serialPort, msg)
        except Exception as e:
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       serialPort, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' %
                            (serialPort, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType,
                   serialPort, failState)
        return failState


class remoteCheck(BaseCheck):
    """A check that receives state from a remote fletch over MQTT"""

    def _runCheck(self):
        thisTime = datetime.now().astimezone(self.timezone)
        timeout = self.getIntOpt('timeout', None)
        failState = self.failState
        et = 0
        if timeout and self.lastUpdate:
            lu = dateparse(self.lastUpdate,
                           tzinfos=LOCALZONES).astimezone(self.timezone)
            et = (thisTime - lu).total_seconds()
            if et > timeout:
                _log.debug('%s (%s): Update timeout %d sec / %s', self.name,
                           self.checkType, et, self.lastUpdate)
                self.log.append('Update timeout %d sec (%s)' %
                                (et, self.lastUpdate))
                failState = True
            else:
                # restore remote log if non-empty
                if self.oldLog:
                    self.log = self.oldLog
        return failState

    def checkData(self, data):
        """Check remote data object for required content"""
        ret = True
        try:
            # int types
            for k in ('threshold', 'failCount'):
                if data[k] is not None:
                    if not isinstance(data[k], int):
                        raise RuntimeError('Invalid %s, expecting int', k)

            # level is optional
            if 'level' not in data:
                data['level'] = None

            # str types
            for k in ('softFail', 'lastCheck', 'lastFail', 'lastPass',
                      'level'):
                if data[k] is not None:
                    if not isinstance(data[k], str):
                        raise RuntimeError('Invalid %s, expecting str', k)

            # failState just needs to be present
            junk = data['failState']

            # log should be an array
            if not isinstance(data['log'], list):
                raise RuntimeError('Invalid log, expecting list')

        except Exception as e:
            _log.warning('%s (%s) Invalid remote data %s: %s', self.name,
                         self.checkType, e.__class__.__name__, e)
            ret = False
        return ret

    def remoteUpdate(self, checkType, data):
        """Report remote transition (replicates baseCheck.update)"""
        if not self.checkData(data):
            _log.warning('%s (%s.%s): Ignored invalid remote data', self.name,
                         self.checkType, checkType)
            return

        self.subType = checkType
        doNotify = False
        if data['failState']:
            if data['failCount'] >= data['threshold']:
                if data['failState'] != self.failState:
                    _log.warning('%s (%s.%s) Log: %r', self.name,
                                 self.checkType, self.subType, data['log'])
                    _log.warning('%s (%s.%s) FAIL', self.name, self.checkType,
                                 self.subType)
                    if self.failAction:
                        doNotify = True
        else:
            if self.failState:
                _log.warning('%s (%s.%s) PASS', self.name, self.checkType,
                             self.subType)
                if self.passAction:
                    doNotify = True

        # Check last update field
        lastUpdate = timeString(self.timezone)
        if 'lastCheck' in data and data['lastCheck']:
            # verify value as a datestring
            try:
                lu = dateparse(data['lastCheck'],
                               tzinfos=LOCALZONES).astimezone(self.timezone)
                lastUpdate = data['lastCheck']
            except Exception:
                _log.info('%s (%s.%s): Ignored invalid last update time',
                          self.name, self.checkType, self.subType)

        # Overwrite state from remote data
        self.failState = data['failState']
        self.lastUpdate = lastUpdate
        self.failCount = data['failCount']
        self.threshold = data['threshold']
        self.log = data['log']
        self.softFail = data['softFail']
        self.lastCheck = data['lastCheck']
        self.lastFail = data['lastFail']
        self.lastPass = data['lastPass']
        self.level = data['level']

        if self.paused:
            _log.debug('%s (%s.%s): Remote check paused', self.name,
                       self.checkType, self.subType)
            self.failState = False
        else:
            if doNotify:
                self.notify()


class diskCheck(BaseCheck):
    """Check a disk volume for free space"""

    # todo: add disk health reporting where available
    def _runCheck(self):
        self.level = None
        volume = self.getStrOpt('volume', '/')
        level = self.getIntOpt('level', defaults.DISKLEVEL)
        hysteresis = self.getIntOpt('hysteresis', defaults.DISKHYSTERESIS)

        failState = True
        try:
            du = disk_usage(volume)
            dpct = 100.0 * du.used / du.total
            if du.total > 0.8 * _TERA:
                msg = '%s: %2.0f%% %0.2f/%0.2fTiB, %0.2fTiB Free, Target: %d%%' % (
                    volume, dpct, du.used / _TERA, du.total / _TERA,
                    du.free / _TERA, level)
            else:
                msg = '%s: %2.0f%% %0.0f/%0.0fGiB, %0.0fGiB Free, Target: %d%%' % (
                    volume, dpct, du.used / _GIGA, du.total / _GIGA,
                    du.free / _GIGA, level)

            self.level = '%2.0f%%' % (dpct, )
            self.log.append(msg)

            prevFail = self.failState
            if prevFail:
                if dpct < (level - hysteresis):  # low trigger
                    failState = False
                else:
                    failState = True
            else:
                if dpct > level:  # high trigger
                    failState = True
                else:
                    failState = False
        except Exception as e:
            _log.debug('%s (%s) %s %s: %s Log=%r', self.name, self.checkType,
                       volume, e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' % (volume, e.__class__.__name__, e))

        _log.debug('%s (%s) %s: Fail=%r', self.name, self.checkType, volume,
                   failState)
        return failState


class cpuCheck(BaseCheck):
    """Check system average cpu usage"""

    def _runCheck(self):
        self.level = None
        level = self.getIntOpt('level', defaults.CPULEVEL)
        hysteresis = self.getIntOpt('hysteresis', defaults.CPUHYSTERESIS)

        failState = True
        try:
            load = cpu_percent()
            if load < 0.001:
                # force an interval for initial data (refer psutil docs)
                _log.debug('%s (%s): CPU load delaying first call', self.name,
                           self.checkType)
                sleep(1.0)
                load = cpu_percent()

            self.level = '%2.0f%%' % (load, )
            msg = 'CPU: %2.0f%%, Target: %d%%' % (load, level)
            self.log.append(msg)

            prevFail = self.failState
            if prevFail:
                if load < (level - hysteresis):  # low trigger
                    failState = False
                else:
                    failState = True
            else:
                if load > level:  # high trigger
                    failState = True
                else:
                    failState = False
        except Exception as e:
            _log.debug('%s (%s) %s: %s Log=%r', self.name, self.checkType,
                       e.__class__.__name__, e, self.log)
            self.log.append('%s: %s' % (e.__class__.__name__, e))

        _log.debug('%s (%s): Fail=%r', self.name, self.checkType, failState)
        return failState


class memoryCheck(BaseCheck):
    """Check system memory usage"""

    def _runCheck(self):
        self.level = None
        level = self.getIntOpt('level', defaults.MEMORYLEVEL)
        hysteresis = self.getIntOpt('hysteresis', defaults.MEMORYHYSTERESIS)

        failState = True
        try:
            mem = virtual_memory()
            if mem.total > 0.8 * _TERA:
                msg = 'Memory: %2.0f%% %0.2f/%0.2fTiB, %0.2fTiB Available, Target: %d%%' % (
                    mem.percent, mem.used / _TERA, mem.total / _TERA,
                    mem.available / _TERA, level)
            else:
                msg = 'Memory: %2.0f%% %0.0f/%0.0fGiB, %0.0fGiB Available, Target: %d%%' % (
                    mem.percent, mem.used / _GIGA, mem.total / _GIGA,
                    mem.available / _GIGA, level)

            self.level = '%2.0f%%' % (mem.percent, )
            self.log.append(msg)

            prevFail = self.failState
            if prevFail:
                if mem.percent < (level - hysteresis):  # low trigger
                    failState = False
                else:
                    failState = True
            else:
                if mem.percent > level:  # high trigger
                    failState = True
                else:
                    failState = False
        except Exception as e:
            _log.debug('%s (%s) %s: %s Log=%r', self.name, self.checkType,
                       e.__class__.__name__, e, self.log)
            self.log.append('%s: %s' % (e.__class__.__name__, e))

        _log.debug('%s (%s): Fail=%r', self.name, self.checkType, failState)
        return failState


class tempCheck(BaseCheck):
    """Network-attached temperature probe check"""

    def _fetchTemp(self, hostname, port, timeout, variant='comet'):
        """Return current temperature reading"""
        h = HTTPConnection(host=hostname, port=port, timeout=timeout)
        h.request('GET', '/event.xml')
        r = h.getresponse()
        tv = None
        if r.status == 200:
            tree = etree.parse(r)
            root = tree.getroot()
            e = root.find('all')
            if e is not None:
                te = e.find('vs1')
                if te is not None:
                    ts = te.text
                    if ts.endswith('\u00b0C'):
                        ts = ts.rstrip('\u00b0C')
                        tv = float(ts)
        else:
            _log.debug('Invalid http response: %r', r.status)

        if tv is None:
            raise RuntimeError('Unable to read temperature from XML')
        return tv

    def _runCheck(self):
        self.level = None
        hostname = self.getStrOpt('hostname', '')
        port = self.getIntOpt('port', 80)
        timeout = self.getIntOpt('timeout', defaults.HTTPSTIMEOUT)
        temperature = self.getIntOpt('temperature', defaults.TEMPLEVEL)
        hysteresis = self.getIntOpt('hysteresis', defaults.TEMPHYSTERESIS)

        failState = True
        try:
            # read current temp
            curTemp = self._fetchTemp(hostname, port, timeout, 'comet')
            prevFail = self.failState

            if prevFail:
                if curTemp < (temperature - hysteresis):  # low trigger
                    failState = False
                else:
                    failState = True
            else:
                if curTemp > temperature:  # high trigger
                    failState = True
                else:
                    failState = False
            msg = '%s: %0.1f\u00b0C Target: %d\u00b0C' % (hostname, curTemp,
                                                          temperature)
            self.level = '%0.1f\u00b0C' % (curTemp, )
            self.log.append(msg)
        except Exception as e:
            _log.debug('%s (%s) %s: %s Log=%r', self.name, self.checkType,
                       e.__class__.__name__, e, self.log)
            self.log.append('%s %s: %s' % (hostname, e.__class__.__name__, e))

        _log.debug('%s (%s): Fail=%r', self.name, self.checkType, failState)
        return failState


class sequenceCheck(BaseCheck):
    """Perform a sequence of checks in turn"""

    def __init__(self, name, options={}):
        super().__init__(name, options)
        self.checks = {}
        self.softFails = set()
        self.levels = {}

    def flatten(self):
        ret = super().flatten()
        ret['options']['checks'] = [c for c in self.checks]
        return ret

    def add_check(self, subcheck):
        """Add check to the sequence"""
        if subcheck is not self:
            self.checks[subcheck.name] = subcheck
            _log.debug('Added check %s to sequence %s', subcheck.name,
                       self.name)

    def del_check(self, name):
        """Remove check from the sequence"""
        if name in self.checks:
            del self.checks[name]
            _log.debug('Removed check %s from sequence %s', name, self.name)

    def replace_check(self, name, subcheck):
        """Replace sequence entry with new check if it existed"""
        if name in self.checks:
            self.del_check(name)
            self.add_check(subcheck)

    def getSummary(self):
        """Return a short summary of failing checks"""
        ret = ''
        if self.failState:
            rv = []
            for check in self.failState.split(','):
                lvl = ''
                mark = '\u26a0\ufe0f'
                if check in self.levels:
                    lvl = self.levels[check]
                if check in self.softFails:
                    mark = '\u26d4'
                rv.append(' %s %s%s' % (check, mark, lvl))
            if rv:
                ret = '\n'.join(rv)
        return ret

    def _runCheck(self):
        self.softfail = set()
        self.levels = {}
        failChecks = set()
        aux = []
        count = 0
        for name in self.checks:
            aux.append((self.checks[name].priority, count, name))
            count += 1
        aux.sort()
        sortedChecks = [n[2] for n in aux]

        # Perform each check in order
        for name in sortedChecks:
            c = self.checks[name]
            cFail = c.update()
            if c.level is not None:
                self.levels[c.name] = c.level
            cMsg = 'PASS'
            if cFail:
                failChecks.add(c.name)
                if c.softFail:
                    self.softFails.add(c.name)
                cMsg = 'FAIL'
                self.log.append('%s (%s): %s' % (c.name, c.checkType, cMsg))
                self.log.extend(c.log)
                self.log.append('')
            else:
                if c.paused:
                    cMsg = 'PAUSED'
                lExtra = ''
                if c.level is not None:
                    lExtra = ' ' + c.level
                self.log.append('%s (%s): %s%s' %
                                (c.name, c.checkType, cMsg, lExtra))

        _log.debug('%s (%s): Fail=%r', self.name, self.checkType, failChecks)

        # Prepare ordered list of failing checks
        rv = []
        for name in sortedChecks:
            if name in failChecks:
                rv.append(name)
        return ','.join(rv)


CHECK_TYPES['cert'] = certCheck
CHECK_TYPES['smtp'] = smtpCheck
CHECK_TYPES['submit'] = submitCheck
CHECK_TYPES['imap'] = imapCheck
CHECK_TYPES['https'] = httpsCheck
CHECK_TYPES['ssh'] = sshCheck
CHECK_TYPES['sequence'] = sequenceCheck
CHECK_TYPES['ups'] = upsStatus
CHECK_TYPES['upstest'] = upsTest
CHECK_TYPES['remote'] = remoteCheck
CHECK_TYPES['disk'] = diskCheck
CHECK_TYPES['memory'] = memoryCheck
CHECK_TYPES['cpu'] = cpuCheck
CHECK_TYPES['temp'] = tempCheck
CHECK_TYPES['dns'] = dnsCheck



# SPDX-License-Identifier: MIT
"""UPS Protocol Support"""

import serial
from logging import getLogger, DEBUG, INFO, WARNING
from time import sleep

_log = getLogger('fletchck.ups')
_log.setLevel(INFO)

# Constants
_ENCODING = 'ascii'


class UpsQsV:
    """UPS 'QS' V Protocol

    Tested with Ninja PSLN600 

    Refer: https://networkupstools.org/protocols/voltronic-qs.html
    """

    def __init__(self, port):
        self._port = serial.Serial(port, 2400, timeout=0.5)
        self.error = False
        self.inputVolts = None
        self.outputVolts = None
        self.load = None
        self.battery = None
        self.fail = False
        self.lowBattery = False
        self.fault = False
        self.testing = False
        self.shutdown = False
        self.beeper = False

    def _write(self, buf):
        _log.debug('SEND: %r', buf)
        return self._port.write(buf.encode(_ENCODING))

    def _read(self, count=1024):
        rcv = self._port.read(count).decode(_ENCODING)
        if rcv == '':
            _log.debug('RECV: [timeout]')
        else:
            _log.debug('RECV: %r', rcv)
        return rcv

    def _command(self, query, responseRequired=True):
        """Send command and wait for reply"""
        self._write(query)
        rcv = self._read()
        if rcv:
            if rcv == query:
                _log.error('Invalid UPS command %r', query)
                rcv = ''
        else:
            if responseRequired:
                _log.debug('Timout waiting for UPS reply')
                self.error = True
        return rcv

    def update(self):
        """Fetch and decode status from connected UPS

        (243.5 245.7 245.4 000 49.8 13.7 --.- 00001000\r
        """
        stat = self._command('QS\r')
        if stat.startswith('('):
            sv = stat[1:-1].split()
            if len(sv) == 8:
                self.inputVolts = float(sv[0])
                self.outputVolts = float(sv[2])
                self.load = int(sv[3])
                self.battery = float(sv[5])
                flags = sv[7]
                self.fail = True if flags[0] == '1' else False
                self.lowBattery = True if flags[1] == '1' else False
                self.fault = True if flags[3] == '1' else False
                self.testing = True if flags[5] == '1' else False
                self.shutdown = True if flags[6] == '1' else False
                self.beeper = True if flags[7] == '1' else False
            else:
                raise RuntimeError('Invalid UPS status: %r' % (stat, ))
        else:
            raise RuntimeError('Invalid UPS status: %r' % (stat, ))
        if self.load > 85:
            _log.warning('High load detected: %d%%', self.load)
        if self.lowBattery:
            _log.warning('Low battery: %0.1fV', self.battery)

    def getInfo(self, update=True):
        if update:
            self.update()
        return 'Fail: %r, Low Battery: %r, Fault: %r, Shutdown: %r, Beeper: %r' % (
            self.fail, self.lowBattery, self.fault, self.shutdown, self.beeper)

    def setShutdown(self, delay=5, recover=5):
        """Shutdown in delay min for recovery after recover minutes of power"""
        if delay < 1:
            delay = 1
        self.update()
        if self.shutdown:
            _log.info('UPS already shutdown')
        cmd = 'S%02dR%04d\r' % (delay, recover)
        return self._command(cmd, responseRequired=False)

    def cancelShutdown(self):
        self.update()
        if not self.shutdown:
            _log.debug('UPS already online')
        self._command('C\r', responseRequired=False)

    def runTest(self):
        """Request a self-test, wait for outcome and return status"""
        self.update()
        if self.fail:
            _log.info('UPS in fail state, test not started')
            return (True, 'UPS in fail state, test not started')
        if self.shutdown:
            _log.info('UPS in shutdown state, test not started')
            return (True, 'UPS in fail state, test not started')
        if self.battery < 13.0:
            _log.info('Low battery %0.1fV, test not started', self.battery)
            return (True, 'Low battery, test not started')
        if self.fault:
            _log.warning('UPS already in fault, test may not pass')

        testLoad = 0
        testVolts = 0.0
        self._command('T\r', responseRequired=False)
        while True:
            self.update()
            if not self.testing:
                break
            if self.load > testLoad:
                testLoad = self.load
            testVolts = self.battery
            _log.debug('Testing: Load=%d%%, Battery=%0.1fV', self.load,
                       self.battery)
            sleep(1)
        _log.debug('Test concluded, Fault: %r', self.fault)
        message = 'Peak load: %d%%, Battery: %0.1fV' % (testLoad, testVolts)
        return (self.fault, message)

    def setBeeper(self, beeper):
        """Set the UPS beeper to the desired value"""
        self.update()
        if self.beeper != beeper:
            self._command('Q\r', responseRequired=False)



# SPDX-License-Identifier: MIT
"""Fletchck application"""

import asyncio
import os.path
from tornado.options import parse_command_line, define, options
from . import util
from . import defaults
from urllib.parse import quote as pathQuote
from logging import getLogger, DEBUG, INFO, WARNING, basicConfig, Formatter
from signal import SIGTERM
from . import mclient

basicConfig(level=DEBUG)
_log = getLogger('fletchck')
_log.setLevel(DEBUG)

# Command line options
define("config", default=None, help="specify site config file", type=str)
define("init", default=False, help="re-initialise system", type=bool)
define("port", default=None, help="specify webui listen port", type=int)
define("merge", default=None, help="merge config from file", type=str)
define("webui", default=True, help="run web ui", type=bool)


class FletchSite():
    """Wrapper object for a single fletchck site instance"""

    def __init__(self):
        self._shutdown = None
        self._lock = asyncio.Lock()
        self._mqtt = None
        self._webui = None
        self._mapCache = None
        self._bgTasks = set()

        self.base = '.'
        self.timezone = None
        self.configFile = defaults.CONFIGPATH
        self.doWebUi = True
        self.webUiPort = None
        self.log = []

        self.scheduler = None
        self.actions = None
        self.checks = None
        self.remotes = {}
        self.remoteTopics = {}
        self.webCfg = None
        self.mqttCfg = None

    def _sigterm(self):
        """Handle TERM signal"""
        _log.warning('Site terminated by SIGTERM')
        self._shutdown.set()

    @classmethod
    def pathQuote(cls, path):
        """URL escape path element for use in link text"""
        return pathQuote(path, safe='')

    def loadConfig(self):
        """Load site from config"""
        util.loadSite(self)

    def testAction(self, actionName):
        """Trigger notification on a single action"""
        ret = True
        if actionName in self.actions:
            _log.warning('Manually notifying action %s', actionName)
            fakeCheck = util.BaseCheck('Test Notification')
            fakeCheck.checkType = 'action-test'
            fakeCheck.failState = True
            fakeCheck.timezone = self.timezone
            fakeCheck.lastFail = util.timeString(self.timezone)
            fakeCheck.lastPass = util.timeString(self.timezone)
            fakeCheck.log = []
            fakeCheck.log.append('Testing action notification')
            fakeCheck.log.append('Fletch version: %s' % (defaults.VERSION))
            val = '%0.0f%%' % (100.0 * util.randbits(8) / 255.0, )
            fakeCheck.log.append('Some random value: %s' % (val))
            fakeCheck.level = val
            ret = True
            _log.warning('Calling trigger on %r', actionName)
            if not self.actions[actionName].trigger(fakeCheck):
                ret = False
        return ret

    def addAction(self, name, config):
        """Add the named action to site"""
        util.addAction(self, name, config)

    def hideOption(self, path, typeName, option):
        """Return a visually-hidden class for form options not in named type"""
        ret = ''
        if path and typeName in defaults.HIDEOPTIONS:
            if option in defaults.HIDEOPTIONS[typeName]:
                ret = ' visually-hidden'

        # override publish when mqtt disabled
        if option == 'publish':
            if not self._mqtt:
                ret = ' visually-hidden'

        return ret

    def copyName(self, checkName):
        """Return a clone check name not already in use on site."""
        if '-copy' in checkName:
            checkName = checkName[0:checkName.rindex('-copy')]
        newName = checkName + '-copy'
        while newName in self.checks:
            newName = '%s-copy-%s' % (checkName, util.token_hex(2))
        return newName

    def checkMap(self, recalculate=False):
        """Return a site map grouped by sorted sequences.

        Orphaned checks (those that do not appear in a sequence)
        are grouped and added to the end of the map with the key
        None.
        """
        if self._mapCache is not None and not recalculate:
            return self._mapCache

        self._mapCache = {}
        seqs = {}
        subchecks = set()
        checkList = self.sortedChecks()

        # collect all defined sequences ## TODO: check recursion (should be ok)
        for checkName in checkList:
            if self.checks[checkName].checkType == 'sequence':
                count = 0
                seqs[checkName] = []
                seq = self.checks[checkName]
                for subCheckName in seq.checks:
                    subchecks.add(subCheckName)
                    sub = self.checks[subCheckName]
                    seqs[checkName].append((sub.priority, count, subCheckName))
                    count += 1

        # reorder sequence subchecks and add to map with priority as value
        for checkName in seqs:
            self._mapCache[checkName] = {}
            seqs[checkName].sort()
            for subcheck in seqs[checkName]:
                self._mapCache[checkName][subcheck[2]] = subcheck[0]

        # add any orphaned checks to map
        self._mapCache[None] = {}
        for checkName in checkList:
            if checkName not in self._mapCache and checkName not in subchecks:
                check = self.checks[checkName]
                self._mapCache[None][checkName] = check.priority

        return self._mapCache

    def sortedChecks(self):
        """Return the list of check names in priority order"""
        aux = []
        count = 0
        for name in self.checks:
            aux.append((self.checks[name].priority, count, name))
            count += 1
        aux.sort()
        return [n[2] for n in aux]

    def addCheck(self, name, config):
        """Add the named check to site"""
        util.addCheck(self, name, config)

    def addRemote(self, name, checkType, remoteId=None):
        """Auto-add a remote check"""
        util.addCheck(self, name, {
            'type': 'remote',
            'subType': checkType,
            'remoteId': remoteId
        })

    def updateCheck(self, name, newName, config):
        """Update existing check to match new config"""
        util.updateCheck(self, name, newName, config)

    def deleteCheck(self, name):
        """Remove a check from a running site"""
        util.deleteCheck(self, name)

    def deleteAction(self, name):
        """Remove an acton from running site"""
        util.deleteAction(self, name)

    def updateUser(self, username, password=None, passHash=None):
        """Update or add user."""
        util.updateUser(self, username, password, passHash)

    def deleteUser(self, username):
        """Remove a user from running site."""
        util.deleteUser(self, username)

    def runCheck(self, name):
        """Run a check by name"""
        if name in self.checks:
            _log.debug('Running check %s', name)
            self.checks[name].update()
            if self.checks[name].publish:
                self.sendMsg(topic=self.checks[name].publish,
                             obj=self.checks[name].msgObj())

    def saveConfig(self):
        """Save site to config"""
        util.saveSite(self)

    def selectConfig(self):
        """Check command line and choose configuration"""
        doStart = True
        parse_command_line()
        if options.config is not None:
            # specify a desired configuration path
            self.configFile = options.config
            self.base = os.path.realpath(os.path.dirname(self.configFile))
        if not options.webui:
            _log.info('Web UI disabled by command line option')
            self.doWebUi = False
        if options.port:
            if isinstance(options.port, int):
                self.webUiPort = max(min(options.port, 65535), 1)
                if options.port != self.webUiPort:
                    _log.warning('Invalid port clamped to %d', self.webUiPort)
            else:
                _log.warning('Ignored invalid web UI port')
        if options.init:
            doStart = util.initSite(self.base, self.doWebUi, self.webUiPort)
        if self.configFile is None:
            self.configFile = defaults.CONFIGPATH
        if options.merge:
            if options.merge.lower().endswith('csv'):
                # assume CSV
                util.mergeCsv(self.base, self.configFile, options.merge)
            else:
                # assume JSON
                util.mergeConfig(self.base, self.configFile, options.merge)
        return doStart

    def getNextRun(self, checkName):
        ret = None
        if checkName in self.checks:
            if self.checks[checkName].trigger is not None:
                check = self.checks[checkName]
                job = self.scheduler.get_job(check.name)
                if job is not None:
                    if job.next_run_time is not None:
                        ret = check.timeString(job.next_run_time)
        return ret

    def getTrigger(self, check):
        return util.trigger2Text(check.trigger)

    def recvMsg(self, topic=None, message=None):
        """MQTT Message receive calback"""
        ob = mclient.fromJson(message)
        if ob is not None and isinstance(ob, dict):
            name = defaults.getOpt('name', ob, str, None)
            remoteId = None
            if name in self.remotes:
                remoteId = name
                name = self.remotes[remoteId]
                _log.debug('Using remoteId=%r for check %r', remoteId, name)
            self.remoteTopics[name] = topic
            checkType = defaults.getOpt('type', ob, str, None)
            data = defaults.getOpt('data', ob, dict, None)
            if name and checkType and data:
                if name not in self.checks:
                    if self.mqttCfg['autoadd']:
                        self.addRemote(name, checkType, remoteId)
                if name in self.checks and self.checks[
                        name].checkType == 'remote':
                    self.checks[name].remoteUpdate(checkType, data)
                else:
                    _log.info('Ignore unconfigured remote check %r', name)
            else:
                _log.info('Ignored malformed MQTT message object')
        else:
            _log.info('Ignored invalid MQTT message object')

    def sendMsg(self, topic, obj):
        """MQTT publish obj to the nominated topic"""
        if self._mqtt is not None:
            self._mqtt.publish_json(topic=topic, obj=obj)

    def clearMsg(self, topic):
        """MQTT publish NULL with retain set to the nominated topic"""
        if self._mqtt is not None:
            _log.debug('Sending NULL to topic=%s', topic)
            self._mqtt.publish(topic=topic, retain=True)

    def getStatus(self):
        status = {
            'fail': False,
            'info': None,
            'checks': {},
            'seqs': {},
            'inseqs': {},
        }
        for checkName in self.checks:
            if checkName not in status['inseqs']:
                status['inseqs'][checkName] = []
            check = self.checks[checkName]
            if check.checkType == 'sequence':
                status['seqs'][checkName] = {}
                for subCheck in check.checks:
                    status['seqs'][checkName][subCheck] = 1
                    if subCheck not in status['inseqs']:
                        status['inseqs'][subCheck] = []
                    status['inseqs'][subCheck].append(checkName)

        failCount = 0
        failCounted = set()
        for checkName in self.sortedChecks():
            check = self.checks[checkName]
            if check.failState:
                if check.checkType != 'sequence':
                    if checkName not in failCounted:
                        failCount += 1
                    failCounted.add(checkName)
                status['fail'] = True
            status['checks'][checkName] = {
                'checkType': check.checkType,
                'failState': check.failState,
                'trigger': check.trigger,
                'softFail': check.softFail if check.softFail else '',
                'lastFail': check.lastFail if check.lastFail else '',
                'lastPass': check.lastPass if check.lastPass else ''
            }
        # re-scan all checks for sequences with failing checks
        # which have not been updated in the sequence itself
        for checkName in self.checks:
            check = self.checks[checkName]
            if check.checkType != 'sequence' and check.failState:
                if checkName in status['inseqs']:
                    for seqName in status['inseqs'][checkName]:
                        if not status['checks'][seqName]['failState']:
                            _log.debug('Sequence %s with failing check: %s',
                                       seqName, checkName)
                            status['checks'][seqName]['failState'] = 'DEP'
        if failCount > 0:
            status['info'] = '%d check%s in fail state' % (
                failCount, 's' if failCount != 1 else '')
        return status

    def reconnectMqtt(self):
        """Replace site MQTT connection"""
        if self._mqtt:
            _log.debug('Disconnecting existing MQTT client')
            self._mqtt.setcb(None)
            self._mqtt.exit()
            self._mqtt = None
        if self.mqttCfg is not None:
            _log.debug('Creating new MQTT client')
            self._mqtt = mclient.Mclient(self.mqttCfg)
            self._mqtt.setcb(self.recvMsg)
            self._mqtt.start()
            if 'basetopic' in self.mqttCfg and self.mqttCfg['basetopic']:
                _log.info('Subscribe basetopic = %s',
                          self.mqttCfg['basetopic'])
                self._mqtt.subscribe(self.mqttCfg['basetopic'])

    def runWebRestart(self):
        """Arrange for a background restart of web UI"""
        task = asyncio.create_task(self.restartWebui())
        self._bgTasks.add(task)
        task.add_done_callback(self._bgTasks.discard)

    async def restartWebui(self):
        """Re-Create web UI"""
        if self._webui is not None:
            await asyncio.sleep(0)
            _log.debug('Stopping existing web server')
            self._webui.stop()
            await self._webui.close_all_connections()
            _log.debug('Web server terminated')
            self._webui = None
        _log.debug('Loading web ui module')
        from . import webui
        self._webui = webui.loadUi(self)

    async def run(self):
        """Load and run site in async loop"""
        rootLogger = getLogger()
        logHandler = util.LogHandler(self)
        logHandler.setLevel(WARNING)
        logHandler.setFormatter(Formatter(defaults.LOGFORMAT))
        rootLogger.addHandler(logHandler)

        self.loadConfig()
        if self.scheduler is None:
            _log.error('Error reading site config')
            return -1

        self._shutdown = asyncio.Event()
        asyncio.get_running_loop().add_signal_handler(SIGTERM, self._sigterm)

        # create mqtt client library handle
        self.reconnectMqtt()

        # create tornado application and listen on configured hostname
        if self.doWebUi and self.webCfg is not None:
            self.runWebRestart()
        else:
            _log.info('Running without webui')

        try:
            _log.warning('Starting')
            await self._shutdown.wait()
            self.saveConfig()
            if self._mqtt:
                self._mqtt.exit()
                self._mqtt.wait()
        except Exception as e:
            _log.error('main %s: %s', e.__class__.__name__, e)

        return 0


def main():
    site = FletchSite()
    if site.selectConfig():
        if site.base and site.base != '.':
            if os.path.exists(site.base):
                os.chdir(site.base)
            else:
                _log.error('Path to site config does not exist')
                return -1
        return asyncio.run(site.run())



# SPDX-License-Identifier: MIT
"""Site defaults"""

# Version String
VERSION = '1.1.0'

# Default application vanity label
APPNAME = 'Fletchck'

# Hostname or IP to listen on
HOSTNAME = 'localhost'

# Fallback default timezone (none = localtime)
TIMEZONE = None

# Start first interval trigger this many seconds after adding
FIRSTTRIGGER = 1

# Random jitter seconds added to execution time of triggers
TRIGGERJITTER = 5

# Configuration filename
CONFIGPATH = 'fletchck.conf'

# SSL cert & key file names, stored in site config
SSLCERT = 'fletchck.cert'
SSLKEY = 'fletchck.key'

# Web UI config skeleton
WEBUICONFIG = {
    'name': APPNAME,
    'hostname': HOSTNAME,
    'port': None,
    'cert': None,
    'key': None,
    'users': None,
}

# MQTT Client Config
MQTTCONFIG = {
    'hostname': 'localhost',
    'port': None,
    'tls': False,
    'username': None,
    'password': None,
    'clientid': None,
    'persist': True,
    'qos': 1,
    'retain': True,
    'basetopic': None,
    'autoadd': True,
    'debug': False,
}

# Format for the volatile log
LOGFORMAT = '%(asctime)s %(message)s'

# Site CSP
CSP = "frame-ancestors 'none'; img-src data: 'self'; default-src 'self'"

# Auth cookie expiry in days
AUTHEXPIRY = 2

# Number of rounds for KDF hash
PASSROUNDS = 16

# Number of random bits in auto-generated passkeys
PASSBITS = 70

# Set of chars to use for auto-generated passkeys
# Note: Only first power of 2 used
PASSCHARS = '0123456789abcdefghjk-pqrst+vwxyz'

# Primary adminsitrator username
ADMINUSER = 'admin'

# SMTP check timeout
SMTPTIMEOUT = 6

# Submit check timeout
SUBMITTIMEOUT = 10

# IMAP check timeout
IMAPTIMEOUT = 6

# HTTPS check timeout
HTTPSTIMEOUT = 10

# SSH check timeout
SSHTIMEOUT = 5

# Certificate check timeout
CERTTIMEOUT = 5

# TLS certificate expiry pre-failure in days
CERTEXPIRYDAYS = 7

# DNS hostname / server
DNSHOSTNAME = '127.0.0.53'

# DNS lifetime
DNSTIMEOUT = 5

# DNS Port
DNSPORT = 53

# DNS over TCP
DNSTCP = False

# Disk full threshold (percent)
DISKLEVEL = 90

# Hysteresis for disk checks (percent)
DISKHYSTERESIS = 2

# Memory usage threshold (percent)
MEMORYLEVEL = 95

# Hysteresis for memory checks (percent)
MEMORYHYSTERESIS = 2

# Average CPU usage threshold (percent)
CPULEVEL = 75

# Hysteresis for cpu checks (percent)
CPUHYSTERESIS = 5

# Temperature warning threshold (degrees C)
TEMPLEVEL = 50

# Hysteresis for temp checks (degrees C)
TEMPHYSTERESIS = 2

# POST Endpoint for SMS Central API
SMSCENTRALURL = 'https://my.smscentral.com.au/api/v3.2'

# POST Endpoint for CK API
CKURL = 'https://sms-api.cloudkinnekt.au/smsp-in'

# Try action trigger this many times before giving up
ACTIONTRIES = 3

# Path to fallback sendmail
SENDMAIL = '/usr/lib/sendmail'

# Hide options for named check types
HIDEOPTIONS = {
    'cert': {
        'remoteId', 'level', 'volume', 'serialPort', 'hostkey', 'reqType',
        'reqPath', 'checks', 'tls', 'beeper', 'temperature', 'hysteresis',
        'reqTcp', 'reqName'
    },
    'submit': {
        'remoteId', 'level', 'volume', 'serialPort', 'hostkey', 'probe',
        'reqType', 'reqPath', 'checks', 'beeper', 'tls', 'temperature',
        'hysteresis', 'reqTcp', 'reqName'
    },
    'smtp': {
        'remoteId', 'level', 'volume', 'serialPort', 'hostkey', 'probe',
        'reqType', 'reqPath', 'checks', 'beeper', 'temperature', 'hysteresis',
        'reqTcp', 'reqName'
    },
    'imap': {
        'remoteId', 'level', 'volume', 'serialPort', 'hostkey', 'probe',
        'reqType', 'reqPath', 'checks', 'beeper', 'tls', 'temperature',
        'hysteresis', 'reqTcp', 'reqName'
    },
    'dns': {
        'remoteId', 'level', 'volume', 'serialPort', 'probe', 'hostkey',
        'reqPath', 'checks', 'beeper', 'tls', 'temperature', 'hysteresis',
        'selfsigned'
    },
    'https': {
        'remoteId', 'level', 'volume', 'serialPort', 'probe', 'hostkey',
        'checks', 'beeper', 'temperature', 'hysteresis', 'reqTcp', 'reqName'
    },
    'ssh': {
        'remoteId', 'level', 'volume', 'serialPort', 'probe', 'reqType',
        'reqPath', 'checks', 'selfsigned', 'tls', 'beeper', 'temperature',
        'hysteresis', 'reqTcp', 'reqName'
    },
    'sequence': {
        'retries', 'remoteId', 'level', 'volume', 'hostname', 'port',
        'serialPort', 'timeout', 'hostkey', 'probe', 'reqType', 'reqPath',
        'selfsigned', 'tls', 'beeper', 'temperature', 'hysteresis', 'reqTcp',
        'reqName'
    },
    'remote': {
        'retries', 'publish', 'level', 'volume', 'hostname', 'port', 'hostkey',
        'probe', 'reqType', 'reqPath', 'checks', 'selfsigned', 'tls',
        'serialPort', 'beeper', 'temperature', 'hysteresis', 'reqTcp',
        'reqName'
    },
    'disk': {
        'remoteId', 'hostname', 'port', 'hostkey', 'probe', 'reqType',
        'reqPath', 'checks', 'selfsigned', 'tls', 'timeout', 'serialPort',
        'beeper', 'temperature', 'reqTcp', 'reqName'
    },
    'cpu': {
        'remoteId',
        'hostname',
        'port',
        'hostkey',
        'probe',
        'reqType',
        'reqPath',
        'checks',
        'selfsigned',
        'tls',
        'timeout',
        'serialPort',
        'beeper',
        'temperature',
        'reqTcp',
        'reqName',
        'volume',
    },
    'memory': {
        'remoteId',
        'hostname',
        'port',
        'hostkey',
        'probe',
        'reqType',
        'reqPath',
        'checks',
        'selfsigned',
        'tls',
        'timeout',
        'serialPort',
        'beeper',
        'temperature',
        'reqTcp',
        'reqName',
        'volume',
    },
    'temp': {
        'remoteId', 'hostkey', 'level', 'volume', 'probe', 'reqType',
        'reqPath', 'checks', 'selfsigned', 'tls', 'serialPort', 'beeper',
        'reqTcp', 'reqName'
    },
    'ups': {
        'remoteId', 'level', 'volume', 'hostname', 'port', 'hostkey', 'probe',
        'reqType', 'reqPath', 'checks', 'selfsigned', 'tls', 'timeout',
        'temperature', 'hysteresis', 'reqTcp', 'reqName'
    },
    'upstest': {
        'remoteId', 'level', 'volume', 'hostname', 'port', 'hostkey', 'probe',
        'reqType', 'reqPath', 'checks', 'selfsigned', 'tls', 'timeout',
        'temperature', 'hysteresis', 'reqTcp', 'reqName'
    },
    'action.email': {
        'url',
        'apikey',
    },
    'action.sms': {
        'site',
        'hostname',
        'port',
        'username',
        'password',
        'timeout',
        'fallback',
    },
    'action.log': {
        'site',
        'url',
        'apikey',
        'sender',
        'recipients',
        'hostname',
        'port',
        'username',
        'password',
        'timeout',
        'fallback',
    },
}


def getOpt(key, store, valType, default=None):
    """Return value of valType from store or default"""
    ret = default
    if key in store and isinstance(store[key], valType):
        ret = store[key]
    return ret
